"""
Installing of certificates to the Windows Certificate Manager
=============================================================

Install certificates to the Windows Certificate Manager

.. code-block:: yaml

    salt://certs/cert.cer:
      certutil.add_store:
        - store: TrustedPublisher
"""

import logging

import salt.utils.platform

log = logging.getLogger(__name__)
__virtualname__ = "certutil"


def __virtual__():
    """
    Only work on Windows
    """
    if salt.utils.platform.is_windows():
        return __virtualname__
    return (False, "Only Windows supported")


def add_store(name, store, saltenv="base"):
    """
    Store a certificate to the given store

    name
        The certificate to store, this can use local paths
        or salt:// paths

    store
        The store to add the certificate to

    saltenv
        The salt environment to use, this is ignored if a local
        path is specified

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    cert_file = __salt__["cp.cache_file"](name, saltenv)
    if cert_file is False:
        ret["result"] = False
        ret["comment"] += "Certificate file not found."
    else:
        cert_serial = __salt__["certutil.get_cert_serial"](cert_file)
        serials = __salt__["certutil.get_stored_cert_serials"](store)

        if cert_serial not in serials:
            retcode = __salt__["certutil.add_store"](name, store, retcode=True)
            if retcode == 0:
                ret["changes"]["added"] = name
            else:
                ret["result"] = False
                ret["comment"] += "Failed to store certificate {}".format(name)
        else:
            ret["comment"] += "{} already stored.".format(name)

    return ret


def del_store(name, store, saltenv="base"):
    """
    Remove a certificate in the given store

    name
        The certificate to remove, this can use local paths
        or salt:// paths

    store
        The store to remove the certificate from

    saltenv
        The salt environment to use, this is ignored if a local
        path is specified

    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    cert_file = __salt__["cp.cache_file"](name, saltenv)
    if cert_file is False:
        ret["result"] = False
        ret["comment"] += "Certificate file not found."
    else:
        cert_serial = __salt__["certutil.get_cert_serial"](cert_file)
        serials = __salt__["certutil.get_stored_cert_serials"](store)

        if cert_serial in serials:
            retcode = __salt__["certutil.del_store"](cert_file, store, retcode=True)
            if retcode == 0:
                ret["changes"]["removed"] = name
            else:
                ret["result"] = False
                ret["comment"] += "Failed to remove the certificate {}".format(name)
        else:
            ret["comment"] += "{} already removed.".format(name)

    return ret



"""
    :codeauthor: Alexander Pyatkin <asp@thexyz.net
"""

import pytest
import salt.utils.json
import salt.utils.path
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.unit import skipIf


@skipIf(salt.utils.path.which("bower") is None, "bower not installed")
class BowerStateTest(ModuleCase, SaltReturnAssertsMixin):
    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_bower_installed_removed(self):
        """
        Basic test to determine if Bower package was successfully installed and
        removed.
        """
        ret = self.run_state("file.directory", name="/salt_test_bower_1", makedirs=True)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "bower.installed", name="underscore", dir="/salt_test_bower_1"
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "bower.removed", name="underscore", dir="/salt_test_bower_1"
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("file.absent", name="/salt_test_bower_1")
        self.assertSaltTrueReturn(ret)

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_bower_installed_pkgs(self):
        """
        Basic test to determine if Bower package successfully installs multiple
        packages.
        """
        ret = self.run_state("file.directory", name="/salt_test_bower_2", makedirs=True)
        self.assertSaltTrueReturn(ret)
        ret = self.run_state(
            "bower.installed",
            name="test",
            dir="/salt_test_bower_2",
            pkgs=["numeral", "underscore"],
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("file.absent", name="/salt_test_bower_2")
        self.assertSaltTrueReturn(ret)

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_bower_installed_from_file(self):
        ret = self.run_state("file.directory", name="/salt_test_bower_3", makedirs=True)
        self.assertSaltTrueReturn(ret)
        bower_json = salt.utils.json.dumps(
            {
                "name": "salt_test_bower_3",
                "dependencies": {"numeral": "~1.5.3", "underscore": "~1.7.0"},
            }
        )
        ret = self.run_state(
            "file.managed", name="/salt_test_bower_3/bower.json", contents=bower_json
        )
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("bower.bootstrap", name="/salt_test_bower_3")
        self.assertSaltTrueReturn(ret)
        ret = self.run_state("file.absent", name="/salt_test_bower_3")
        self.assertSaltTrueReturn(ret)



"""
Dot NET functions

.. versionadded:: 3001
"""

import salt.utils.platform
import salt.utils.win_reg as win_reg
from salt.utils.versions import LooseVersion

__virtualname__ = "dotnet"


# Although utils are often directly imported, it is also possible to use the
# loader.
def __virtual__():
    """
    Only load if platform is Windows
    """
    if not salt.utils.platform.is_windows():
        return False, "This utility only works on Windows"

    return __virtualname__


def versions():
    """
    Figure out what versions of .NET are installed on the system

    Returns:
        dict: A dictionary containing two keys:
            - versions: A list of versions installed on the system
            - details: A dictionary with details about the versions installed on
              the system
    """
    hive = "HKLM"
    key = "SOFTWARE\\Microsoft\\NET Framework Setup\\NDP"
    ver_keys = win_reg.list_keys(hive=hive, key=key)

    def dotnet_45_plus_versions(release):
        if release >= 528040:
            return "4.8"
        if release >= 461808:
            return "4.7.2"
        if release >= 461308:
            return "4.7.1"
        if release >= 460798:
            return "4.7"
        if release >= 394802:
            return "4.6.2"
        if release >= 394254:
            return "4.6.1"
        if release >= 393295:
            return "4.6"
        if release >= 379893:
            return "4.5.2"
        if release >= 378675:
            return "4.5.1"
        if release >= 378389:
            return "4.5"

    return_dict = {"versions": [], "details": {}}
    for ver_key in ver_keys:

        if ver_key.startswith("v"):
            if win_reg.value_exists(
                hive=hive, key="\\".join([key, ver_key]), vname="Version"
            ):
                # https://docs.microsoft.com/en-us/dotnet/framework/migration-guide/how-to-determine-which-versions-are-installed#find-net-framework-versions-1-4-with-codep
                install = win_reg.read_value(
                    hive=hive, key="\\".join([key, ver_key]), vname="Install"
                )["vdata"]
                if not install:
                    continue
                version = win_reg.read_value(
                    hive=hive, key="\\".join([key, ver_key]), vname="Version"
                )["vdata"]
                sp = win_reg.read_value(
                    hive=hive, key="\\".join([key, ver_key]), vname="SP"
                )["vdata"]
            elif win_reg.value_exists(
                hive=hive, key="\\".join([key, ver_key, "Full"]), vname="Release"
            ):
                # https://docs.microsoft.com/en-us/dotnet/framework/migration-guide/how-to-determine-which-versions-are-installed#find-net-framework-versions-45-and-later-with-code
                install = win_reg.read_value(
                    hive=hive, key="\\".join([key, ver_key, "Full"]), vname="Install"
                )["vdata"]
                if not install:
                    continue
                version = dotnet_45_plus_versions(
                    win_reg.read_value(
                        hive=hive,
                        key="\\".join([key, ver_key, "Full"]),
                        vname="Release",
                    )["vdata"]
                )
                sp = "N/A"
            else:
                continue

            service_pack = " SP{}".format(sp) if sp != "N/A" else ""
            return_dict["versions"].append(version)
            return_dict["details"][ver_key] = {
                "version": version,
                "service_pack": sp,
                "full": "{}{}".format(version, service_pack),
            }

    return return_dict


def versions_list():
    """
    Get a sorted list of .NET versions installed on the system

    Returns:
        list: A sorted list of versions installed on the system
    """
    return sorted(versions()["versions"])


def versions_details():
    """
    Get the details for all versions of .NET installed on a system

    Returns:
        dict: A dictionary of details for each version on the system. Contains
        the following keys:
            - version: The version installed
            - service_pack: The service pack for the version installed
            - full: The full version name including the service pack
    """
    return versions()["details"]


def version_at_least(version):
    """
    Check that the system contains a version of .NET that is at least the
    passed version.

    Args:

        version (str): The version to check for

    Returns:
        bool: ``True`` if the system contains a version of .NET that is at least
        the passed version, otherwise ``False``
    """
    for dotnet_version in versions_list():
        if LooseVersion(dotnet_version) >= LooseVersion(str(version)):
            return True
    return False



"""
Manage RabbitMQ Upstreams
=========================

Example:

.. code-block:: yaml

    rabbit_upstream:
      rabbitmq_upstream.present:
      - name: upstream_1
      - uri: amqp://my_user:my_password@rabbitmq_host
      - trust_user_id: True
      - ack_mode: on-confirm
      - max_hops: 1

.. versionadded:: 3000
"""


import json
import logging

import salt.utils.data
import salt.utils.dictdiffer
from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if the appropriate rabbitmq module functions are loaded.
    """
    requirements = [
        "rabbitmq.list_upstreams",
        "rabbitmq.upstream_exists",
        "rabbitmq.set_upstream",
        "rabbitmq.delete_upstream",
    ]
    if all(req in __salt__ for req in requirements):
        return True
    return (False, "rabbitmq module could not be loaded")


def present(
    name,
    uri,
    prefetch_count=None,
    reconnect_delay=None,
    ack_mode=None,
    trust_user_id=None,
    exchange=None,
    max_hops=None,
    expires=None,
    message_ttl=None,
    ha_policy=None,
    queue=None,
    runas=None,
):
    """
    Ensure the RabbitMQ upstream exists.

    :param str name: The name of the upstream connection
    :param str uri: The URI to connect to. If upstream is a cluster and can have
        several URIs, you can enter them here separated by spaces.
        Examples:
        - amqp://user:password@server_name
        - amqp://user:password@server_name/vhost
        When connecting with SSL, several URI-parameters need also be specified:
        - cacertfile = /path/to/cacert.pem
        - certfile = /path/to/cert.pem
        - keyfile = /part/to/key.pem
        - verity = verify_peer
        - fail_if_no_peer_cert = true | false
        - auth_mechanism = external
        Example:
        - amqp://user:password@server_name?cacertfile=/path/to/cacert.pem&\
            certfile=/path/to/cert.pem&keyfile=/path/to/key.pem&verify=verify_peer
        - amqp://server-name?cacertfile=/path/to/cacert.pem&certfile=/path/to/cert.pem&\
            keyfile=/path/to/key.pem&verify=verify_peer&fail_if_no_peer_cert=true&auth_mechanism=external
    :param int prefetch_count: Maximum number of unacknowledged messages that may
        be in flight over a federation link at one time. Default: 1000
    :param int reconnect_delay: Time in seconds to wait after a network link
        goes down before attempting reconnection. Default: 5
    :param str ack_mode: The following values are allowed:
        on-confirm: Messages are acknowledged to the upstream broker after they
        have been confirmed downstream. Handles network errors and broker failures
        without losing messages. The slowest option, and the default.
        on-publish: Messages are acknowledged to the upstream broker after they
        have been published downstream. Handles network errors without losing
        messages, but may lose messages in the event of broker failures.
        no-ack: Message acknowledgements are not used. The fastest option, but
        you may lose messages in the event of network or broker failures.
    :param bool trust_user_id: Set ``True`` to preserve the "user-id" field across
        a federation link, even if the user-id does not match that used to republish
        the message. Set to ``False`` to clear the "user-id" field when messages
        are federated. Only set this to ``True`` if you trust the upstream broker
        not to forge user-ids.
    :param str exchange: The name of the upstream exchange. Default is to use the
        same name as the federated exchange.
    :param int max_hops: Maximum number of federation links that messages can
        traverse before being dropped. Defaults to 1 if not set.
    :param int expires: Time in milliseconds that the upstream should remember
        about this node for. After this time all upstream state will be removed.
        Set to ``None`` (Default) to mean "forever".
    :param int message_ttl: Time in milliseconds that undelivered messages should
        be held upstream when there is a network outage or backlog.
        Set to ``None`` (default) to mean "forever".
    :param str ha_policy: Determines the "x-ha-policy"-argument for the upstream
        queue for a federated exchange. Default is "none" meaning the queue is
        not HA.
    :param str queue: The name of the upstream queue. Default is to use the same
        name as the federated queue.

    .. versionadded:: 3000

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}
    action = None

    try:
        current_upstreams = __salt__["rabbitmq.list_upstreams"](runas=runas)
    except CommandExecutionError as err:
        ret["comment"] = "Error: {}".format(err)
        return ret
    new_config = salt.utils.data.filter_falsey(
        {
            "uri": uri,
            "prefetch-count": prefetch_count,
            "reconnect-delay": reconnect_delay,
            "ack-mode": ack_mode,
            "trust-user-id": trust_user_id,
            "exchange": exchange,
            "max-hops": max_hops,
            "expires": expires,
            "message-ttl": message_ttl,
            "ha-policy": ha_policy,
            "queue": queue,
        }
    )

    if name in current_upstreams:
        current_config = json.loads(current_upstreams.get(name, ""))
        diff_config = salt.utils.dictdiffer.deep_diff(current_config, new_config)
        if diff_config:
            action = "update"
        else:
            ret["result"] = True
            ret["comment"] = 'Upstream "{}" already present as specified.'.format(name)
    else:
        action = "create"
        diff_config = {"old": None, "new": new_config}

    if action:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = 'Upstream "{}" would have been {}d.'.format(name, action)
        else:
            try:
                res = __salt__["rabbitmq.set_upstream"](
                    name,
                    uri,
                    prefetch_count=prefetch_count,
                    reconnect_delay=reconnect_delay,
                    ack_mode=ack_mode,
                    trust_user_id=trust_user_id,
                    exchange=exchange,
                    max_hops=max_hops,
                    expires=expires,
                    message_ttl=message_ttl,
                    ha_policy=ha_policy,
                    queue=queue,
                    runas=runas,
                )
                ret["result"] = res
                ret["comment"] = 'Upstream "{}" {}d.'.format(name, action)
                ret["changes"] = diff_config
            except CommandExecutionError as exp:
                ret["comment"] = "Error trying to {} upstream: {}".format(action, exp)
    return ret


def absent(name, runas=None):
    """
    Ensure the named upstream is absent.

    :param str name: The name of the upstream to remove
    :param str runas: User to run the command

    .. versionadded:: 3000
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    try:
        upstream_exists = __salt__["rabbitmq.upstream_exists"](name, runas=runas)
    except CommandExecutionError as err:
        ret["comment"] = "Error: {}".format(err)
        return ret

    if upstream_exists:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = 'Upstream "{}" would have been deleted.'.format(name)
        else:
            try:
                res = __salt__["rabbitmq.delete_upstream"](name, runas=runas)
                if res:
                    ret["result"] = True
                    ret["comment"] = 'Upstream "{}" has been deleted.'.format(name)
                    ret["changes"] = {"old": name, "new": None}
            except CommandExecutionError as err:
                ret["comment"] = "Error: {}".format(err)
    else:
        ret["result"] = True
        ret["comment"] = 'The upstream "{}" is already absent.'.format(name)
    return ret



import pytest
import salt.config
import salt.loader
import salt.modules.saltutil
import salt.state
from tests.support.mock import patch


@pytest.fixture
def opts(salt_master_factory):
    config_overrides = {"master_uri": "tcp://127.0.0.1:11111"}
    factory = salt_master_factory.salt_minion_daemon(
        "get-tops-minion",
        overrides=config_overrides,
    )
    yield factory.config.copy()


@pytest.fixture
def modules(opts):
    yield salt.loader.minion_mods(opts, context={})


@pytest.fixture
def configure_mocks(opts):
    with patch("salt.utils.extmods.sync", return_value=(None, None)):
        with patch.object(salt.state.HighState, "top_matches", return_value={}):
            # Mock the __gen_opts method of HighState so it doesn't try to auth to master.
            with patch.object(
                salt.state.BaseHighState, "_BaseHighState__gen_opts", return_value=opts
            ):
                # Mock the _gather_pillar method of State so it doesn't try to auth to master.
                with patch.object(salt.state.State, "_gather_pillar", return_value={}):
                    yield


@pytest.fixture
def destroy(configure_mocks):
    with patch.object(salt.state.HighState, "destroy") as destroy:
        yield destroy


@pytest.fixture
def get_top(configure_mocks):
    with patch.object(salt.state.HighState, "get_top") as get_top:
        yield get_top


def test__get_top_file_envs(modules, get_top, destroy):
    """
    Ensure we cleanup objects created by saltutil._get_top_file_envs #60449
    """
    modules["saltutil.sync_clouds"]()
    assert get_top.called
    # Ensure destroy is getting called
    assert destroy.called



import re

import pytest
from tests.support.case import ModuleCase


@pytest.mark.windows_whitelisted
class KeyModuleTest(ModuleCase):
    @pytest.mark.slow_test
    def test_key_finger(self):
        """
        test key.finger to ensure we receive a valid fingerprint
        """
        out = self.run_function("key.finger")
        match = re.match("([0-9a-z]{2}:){15,}[0-9a-z]{2}$", out)
        self.assertTrue(match)

    @pytest.mark.slow_test
    def test_key_finger_master(self):
        """
        test key.finger_master to ensure we receive a valid fingerprint
        """
        out = self.run_function("key.finger_master")
        match = re.match("([0-9a-z]{2}:){15,}[0-9a-z]{2}$", out)
        self.assertTrue(match)



import fnmatch

import salt.modules.config as config
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase

DEFAULTS = {
    "test.option.foo": "value of test.option.foo in DEFAULTS",
    "test.option.bar": "value of test.option.bar in DEFAULTS",
    "test.option.baz": "value of test.option.baz in DEFAULTS",
    "test.option": "value of test.option in DEFAULTS",
}


class TestModulesConfig(TestCase, LoaderModuleMockMixin):

    no_match = "test.option.nope"
    opt_name = "test.option.foo"
    wildcard_opt_name = "test.option.b*"

    def setup_loader_modules(self):
        return {
            config: {
                "__opts__": {
                    "test.option.foo": "value of test.option.foo in __opts__",
                    "test.option.bar": "value of test.option.bar in __opts__",
                    "test.option.baz": "value of test.option.baz in __opts__",
                },
                "__pillar__": {
                    "test.option.foo": "value of test.option.foo in __pillar__",
                    "test.option.bar": "value of test.option.bar in __pillar__",
                    "test.option.baz": "value of test.option.baz in __pillar__",
                    "master": {
                        "test.option.foo": "value of test.option.foo in master",
                        "test.option.bar": "value of test.option.bar in master",
                        "test.option.baz": "value of test.option.baz in master",
                    },
                },
                "__grains__": {
                    "test.option.foo": "value of test.option.foo in __grains__",
                    "test.option.bar": "value of test.option.bar in __grains__",
                    "test.option.baz": "value of test.option.baz in __grains__",
                },
            }
        }

    def _wildcard_match(self, data):
        return {x: data[x] for x in fnmatch.filter(data, self.wildcard_opt_name)}

    def test_defaults_only_name(self):
        with patch.dict(config.DEFAULTS, DEFAULTS):
            opt_name = "test.option"
            opt = config.option(opt_name)
            self.assertEqual(opt, config.DEFAULTS[opt_name])

    def test_no_match(self):
        """
        Make sure that the defa
        """
        with patch.dict(config.DEFAULTS, DEFAULTS):
            ret = config.option(self.no_match)
            assert ret == "", ret

            default = "wat"
            ret = config.option(self.no_match, default=default)
            assert ret == default, ret

            ret = config.option(self.no_match, wildcard=True)
            assert ret == {}, ret

            default = {"foo": "bar"}
            ret = config.option(self.no_match, default=default, wildcard=True)
            assert ret == default, ret

            # Should be no match since wildcard=False
            ret = config.option(self.wildcard_opt_name)
            assert ret == "", ret

    def test_omits(self):
        with patch.dict(config.DEFAULTS, DEFAULTS):

            # ********** OMIT NOTHING **********

            # Match should be in __opts__ dict
            ret = config.option(self.opt_name)
            assert ret == config.__opts__[self.opt_name], ret

            # Wildcard match
            ret = config.option(self.wildcard_opt_name, wildcard=True)
            assert ret == self._wildcard_match(config.__opts__), ret

            # ********** OMIT __opts__ **********

            # Match should be in __grains__ dict
            ret = config.option(self.opt_name, omit_opts=True)
            assert ret == config.__grains__[self.opt_name], ret

            # Wildcard match
            ret = config.option(self.wildcard_opt_name, omit_opts=True, wildcard=True)
            assert ret == self._wildcard_match(config.__grains__), ret

            # ********** OMIT __opts__, __grains__ **********

            # Match should be in __pillar__ dict
            ret = config.option(self.opt_name, omit_opts=True, omit_grains=True)
            assert ret == config.__pillar__[self.opt_name], ret

            # Wildcard match
            ret = config.option(
                self.wildcard_opt_name, omit_opts=True, omit_grains=True, wildcard=True
            )
            assert ret == self._wildcard_match(config.__pillar__), ret

            # ********** OMIT __opts__, __grains__, __pillar__ **********

            # Match should be in master opts
            ret = config.option(
                self.opt_name, omit_opts=True, omit_grains=True, omit_pillar=True
            )
            assert ret == config.__pillar__["master"][self.opt_name], ret

            # Wildcard match
            ret = config.option(
                self.wildcard_opt_name,
                omit_opts=True,
                omit_grains=True,
                omit_pillar=True,
                wildcard=True,
            )
            assert ret == self._wildcard_match(config.__pillar__["master"]), ret

            # ********** OMIT ALL THE THINGS **********

            # Match should be in master opts
            ret = config.option(
                self.opt_name,
                omit_opts=True,
                omit_grains=True,
                omit_pillar=True,
                omit_master=True,
            )
            assert ret == config.DEFAULTS[self.opt_name], ret

            # Wildcard match
            ret = config.option(
                self.wildcard_opt_name,
                omit_opts=True,
                omit_grains=True,
                omit_pillar=True,
                omit_master=True,
                wildcard=True,
            )
            assert ret == self._wildcard_match(config.DEFAULTS), ret

            # Match should be in master opts
            ret = config.option(self.opt_name, omit_all=True)
            assert ret == config.DEFAULTS[self.opt_name], ret

            # Wildcard match
            ret = config.option(self.wildcard_opt_name, omit_all=True, wildcard=True)
            assert ret == self._wildcard_match(config.DEFAULTS), ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Taken from sphinx-contrib
# https://bitbucket.org/birkenfeld/sphinx-contrib/src/a3d904f8ab24/youtube

# If not otherwise noted, the extensions in this package are licensed
# under the following license.
#
# Copyright (c) 2009 by the contributors (see AUTHORS file).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import division

import re

from docutils import nodes
from docutils.parsers.rst import directives

try:
    from sphinx.util.compat import Directive
except ImportError:
    from docutils.parsers.rst import Directive

CONTROL_HEIGHT = 30


def get_size(d, key):
    if key not in d:
        return None
    m = re.match("(\d+)(|%|px)$", d[key])
    if not m:
        raise ValueError("invalid size %r" % d[key])
    return int(m.group(1)), m.group(2) or "px"


def css(d):
    return "; ".join(sorted("%s: %s" % kv for kv in d.iteritems()))


class youtube(nodes.General, nodes.Element):
    pass


def visit_youtube_node(self, node):
    aspect = node["aspect"]
    width = node["width"]
    height = node["height"]

    if aspect is None:
        aspect = 16, 9

    if (height is None) and (width is not None) and (width[1] == "%"):
        style = {
            "padding-top": "%dpx" % CONTROL_HEIGHT,
            "padding-bottom": "%f%%" % (width[0] * aspect[1] / aspect[0]),
            "width": "%d%s" % width,
            "position": "relative",
        }
        self.body.append(self.starttag(node, "div", style=css(style)))
        style = {
            "position": "absolute",
            "top": "0",
            "left": "0",
            "width": "100%",
            "height": "100%",
            "border": "0",
        }
        attrs = {
            "src": "http://www.youtube.com/embed/%s" % node["id"],
            "style": css(style),
        }
        self.body.append(self.starttag(node, "iframe", **attrs))
        self.body.append("</iframe></div>")
    else:
        if width is None:
            if height is None:
                width = 560, "px"
            else:
                width = height[0] * aspect[0] / aspect[1], "px"
        if height is None:
            height = width[0] * aspect[1] / aspect[0], "px"
        style = {
            "width": "%d%s" % width,
            "height": "%d%s" % (height[0] + CONTROL_HEIGHT, height[1]),
            "border": "0",
        }
        attrs = {
            "src": "http://www.youtube.com/embed/%s" % node["id"],
            "style": css(style),
        }
        self.body.append(self.starttag(node, "iframe", **attrs))
        self.body.append("</iframe>")


def depart_youtube_node(self, node):
    pass


class YouTube(Directive):
    has_content = True
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {
        "width": directives.unchanged,
        "height": directives.unchanged,
        "aspect": directives.unchanged,
    }

    def run(self):
        if "aspect" in self.options:
            aspect = self.options.get("aspect")
            m = re.match("(\d+):(\d+)", aspect)
            if m is None:
                raise ValueError("invalid aspect ratio %r" % aspect)
            aspect = tuple(int(x) for x in m.groups())
        else:
            aspect = None
        width = get_size(self.options, "width")
        height = get_size(self.options, "height")
        return [
            youtube(id=self.arguments[0], aspect=aspect, width=width, height=height)
        ]


def setup(app):
    app.add_node(youtube, html=(visit_youtube_node, depart_youtube_node))
    app.add_directive("youtube", YouTube)



"""
A module that adds data to the Pillar structure from a NetBox API.

.. versionadded:: 2019.2.0

Configuring the NetBox ext_pillar
---------------------------------

To use this pillar, you must first create a token in your NetBox instance at
http://netbox.example.com/user/api-tokens/ (substituting the hostname of your
NetBox instance)

The NetBox api_url and api_token must be set in the master
config.

For example ``/etc/salt/master.d/netbox.conf``:

.. code-block:: yaml

  ext_pillar:
    - netbox:
        api_url: http://netbox.example.com/api/
        api_token: 123abc


The following options are optional, and determine whether or not
the module will attempt to configure the ``proxy`` pillar data for
use with the napalm proxy-minion:

.. code-block:: yaml

  proxy_return: True
  proxy_username: admin

By default, this module will query the NetBox API for the platform
associated with the device, and use the 'NAPALM driver' field to
set the napalm proxy-minion driver. (Currently only 'napalm' is supported
for drivertype.)

This module currently only supports the napalm proxy minion and assumes
you will use SSH keys to authenticate to the network device.  If password
authentication is desired, it is recommended to create another ``proxy``
key in pillar_roots (or git_pillar) with just the ``passwd`` key and use
:py:func:`salt.renderers.gpg <salt.renderers.gpg>` to encrypt the value.

If you use more than one username for your devices, leave proxy_username unset,
and set the ``username`` key in your pillar as well. If any additional options
for the proxy setup are needed, they should also be configured in pillar_roots.

Other available configuration options:

site_details: ``True``
    Whether should retrieve details of the site the device belongs to.

site_prefixes: ``True``
    Whether should retrieve the prefixes of the site the device belongs to.

devices: ``True``
    .. versionadded:: 3004.0

    Whether should retrieve physical devices.

virtual_machines: ``False``
    .. versionadded:: 3004.0

    Whether should retrieve virtual machines.

interfaces: ``False``
    .. versionadded:: 3004.0

    Whether should retrieve the interfaces of the device.

interface_ips: ``False``
    .. versionadded:: 3004.0

    Whether should retrieve the IP addresses for interfaces of the device.
    (interfaces must be set to True as well)

api_query_result_limit: ``Use NetBox default``
    .. versionadded:: 3004.0

    An integer specifying how many results should be returned for each query
    to the NetBox API. Leaving this unset will use NetBox's default value.

Note that each option you enable can have a detrimental impact on pillar
performance, so use them with caution.

After configuring the pillar, you must restart the Salt master for the changes
to take effect.

For example:

.. code-block:: shell

  systemctl restart salt-master

To query perform a quick test of the pillar, you should refresh the pillar on
the minion with the following:

.. code-block:: shell

  salt minion1 saltutil.refresh_pillar

And then query the pillar:

.. code-block:: shell

  salt minion1 pillar.items 'netbox'

Example output:

.. code-block:: text

  minion1:
      netbox:
          ----------
          id:
              511
          url:
              https://netbox.example.com/api/dcim/devices/511/
          name:
              minion1
          node_type:
              device
          display_name:
              minion1
          device_type:
              ----------
              id:
                  4
              url:
                  https://netbox.example.com/api/dcim/device-types/4/
              manufacturer:
                  ----------
                  id:
                      1
                  url:
                      https://netbox.example.com/api/dcim/manufacturers/1/
                  name:
                      Cisco
                  slug:
                      cisco
              model:
                  ISR2901
              slug:
                  isr2901
              display_name:
                  Cisco ISR2901
          device_role:
              ----------
              id:
                  45
              url:
                  https://netbox.example.com/api/dcim/device-roles/45/
              name:
                  Network
              slug:
                  network
          interfaces:
              |_
                ----------
                id:
                    8158
                ip_addresses:
                    |_
                      ----------
                      id:
                          1146
                      url:
                          https://netbox.example.com/api/ipam/ip-addresses/1146/
                      family:
                          ----------
                          value:
                              4
                          label:
                              IPv4
                      address:
                          192.0.2.1/24
                      vrf:
                          None
                      tenant:
                          None
                      status:
                          ----------
                          value:
                              active
                           label:
                              Active
                      role:
                          None
                      nat_inside:
                          None
                      nat_outside:
                          None
                      dns_name:
                      description:
                      tags:
                      custom_fields:
                      created:
                          2021-02-19
                      last_updated:
                          2021-02-19T06:12:04.153386Z
                url:
                    https://netbox.example.com/api/dcim/interfaces/8158/
                name:
                    GigabitEthernet0/0
                label:
                type:
                    ----------
                    value:
                        1000base-t
                    label:
                        1000BASE-T (1GE)
                enabled:
                    True
                lag:
                    None
                mtu:
                    None
                mac_address:
                    None
                mgmt_only:
                    False
                description:
                mode:
                    None
                untagged_vlan:
                    None
                tagged_vlans:
                cable:
                    None
                cable_peer:
                    None
                cable_peer_type:
                    None
                connected_endpoint:
                    None
                connected_endpoint_type:
                    None
                connected_endpoint_reachable:
                    None
                tags:
                count_ipaddresses:
                    1
              |_
                ----------
                id:
                    8159
                ip_addresses:
                    |_
                      ----------
                      id:
                          1147
                      url:
                          https://netbox.example.com/api/ipam/ip-addresses/1147/
                      family:
                          ----------
                          value:
                              4
                          label:
                              IPv4
                      address:
                          198.51.100.1/24
                      vrf:
                          None
                      tenant:
                          None
                      status:
                          ----------
                          value:
                              active
                          label:
                              Active
                      role:
                          None
                      nat_inside:
                          None
                      nat_outside:
                          None
                      dns_name:
                      description:
                      tags:
                      custom_fields:
                      created:
                          2021-02-19
                      last_updated:
                          2021-02-19T06:12:40.508154Z
                      url:
                          https://netbox.example.com/api/dcim/interfaces/8159/
                      name:
                          GigabitEthernet0/1
                      label:
                      type:
                          ----------
                          value:
                              1000base-t
                          label:
                              1000BASE-T (1GE)
                      enabled:
                          True
                      lag:
                          None
                      mtu:
                          None
                      mac_address:
                          None
                      mgmt_only:
                          False
                      description:
                      mode:
                          None
                      untagged_vlan:
                          None
                      tagged_vlans:
                      cable:
                          None
                      cable_peer:
                          None
                      cable_peer_type:
                          None
                      connected_endpoint:
                          None
                      connected_endpoint_type:
                          None
                      connected_endpoint_reachable:
                          None
                      tags:
                      count_ipaddresses:
                          1
          tenant:
              None
          platform:
              ----------
              id:
                  1
              url:
                  https://netbox.example.com/api/dcim/platforms/1/
              name:
                  Cisco IOS
              slug:
                  ios
          serial:
          asset_tag:
              None
          site:
              ----------
              id:
                  18
              url:
                  https://netbox.example.com/api/dcim/sites/18/
              name:
                  Site 1
              slug:
                  site1
              status:
                  ----------
                  value:
                      active
                  label:
                      Active
              region:
                  None
              tenant:
                  None
              facility:
              asn:
                  None
              time_zone:
                  None
              description:
              physical_address:
              shipping_address:
              latitude:
                  None
              longitude:
                  None
              contact_name:
              contact_phone:
              contact_email:
              comments:
              tags:
              custom_fields:
              created:
                  2021-02-25
              last_updated:
                  2021-02-25T14:21:07.898957Z
              circuit_count:
                  0
              device_count:
                  1
              prefix_count:
                  2
              rack_count:
                  0
              virtualmachine_count:
                  1
              vlan_count:
                  0
              prefixes:
                  |_
                    ----------
                    id:
                        284
                    url:
                        https://netbox.example.com/api/ipam/prefixes/284/
                    family:
                        ----------
                        value:
                            4
                        label:
                            IPv4
                    prefix:
                        192.0.2.0/24
                    vrf:
                        None
                    tenant:
                        None
                    vlan:
                        None
                          ----------
                        value:
                            active
                        label:
                            Active
                    role:
                        None
                    is_pool:
                        False
                    description:
                    tags:
                    custom_fields:
                    created:
                        2021-02-25
                    last_updated:
                        2021-02-25T15:08:27.136305Z
                  |_
                    ----------
                    id:
                        285
                    url:
                        https://netbox.example.com/api/ipam/prefixes/285/
                    family:
                        ----------
                        value:
                            4
                        label:
                            IPv4
                    prefix:
                        198.51.100.0/24
                    vrf:
                        None
                    tenant:
                        None
                    vlan:
                        None
                    status:
                        ----------
                        value:
                            active
                        label:
                            Active
                    role:
                        None
                    is_pool:
                        False
                    description:
                    tags:
                    custom_fields:
                    created:
                        2021-02-25
                    last_updated:
                        2021-02-25T15:08:59.880440Z
          rack:
              None
          position:
              None
          face:
              None
          parent_device:
              None
          status:
              ----------
              value:
                  active
              label:
                  Active
          primary_ip:
              ----------
              id:
                  1146
              url:
                  https://netbox.example.com/api/ipam/ip-addresses/1146/
              family:
                  4
              address:
                  192.0.2.1/24
          primary_ip4:
              ----------
              id:
                  1146
              url:
                  https://netbox.example.com/api/ipam/ip-addresses/1146/
              family:
                  4
              address:
                  192.0.2.1/24
          primary_ip6:
              None
          cluster:
              None
          virtual_chassis:
              None
          vc_position:
              None
          vc_priority:
              None
          comments:
          local_context_data:
              None
          tags:
          custom_fields:
          config_context:
          created:
              2021-02-19
          last_updated:
              2021-02-19T06:12:04.171105Z
"""

import logging

import salt.utils.http
import salt.utils.url
from salt._compat import ipaddress

# Set up logging
log = logging.getLogger(__name__)


def _get_devices(api_url, minion_id, headers, api_query_result_limit):
    device_url = "{api_url}/{app}/{endpoint}".format(
        api_url=api_url, app="dcim", endpoint="devices"
    )
    device_results = []
    params = {"name": minion_id}
    if api_query_result_limit:
        params["limit"] = api_query_result_limit
    device_ret = salt.utils.http.query(
        device_url, params=params, header_dict=headers, decode=True
    )
    while True:
        # Check status code for API call
        if "error" in device_ret:
            log.error(
                'API query failed for "%s", status code: %d, error %s',
                minion_id,
                device_ret["status"],
                device_ret["error"],
            )
            return []
        else:
            device_results.extend(device_ret["dict"]["results"])
        # Check if we need to paginate and fetch the next result list
        if device_ret["dict"]["next"]:
            device_ret = salt.utils.http.query(
                device_ret["dict"]["next"], header_dict=headers, decode=True
            )
        else:
            break

    # Set the node type
    device_count = 0
    for device in device_results:
        device_results[device_count]["node_type"] = "device"
        device_count += 1

    # Return the results
    return device_results


def _get_virtual_machines(api_url, minion_id, headers, api_query_result_limit):
    vm_url = "{api_url}/{app}/{endpoint}".format(
        api_url=api_url, app="virtualization", endpoint="virtual-machines"
    )
    vm_results = []
    params = {"name": minion_id}
    if api_query_result_limit:
        params["limit"] = api_query_result_limit
    vm_ret = salt.utils.http.query(
        vm_url, params=params, header_dict=headers, decode=True
    )
    while True:
        # Check status code for API call
        if "error" in vm_ret:
            log.error(
                'API query failed for "%s", status code: %d, error %s',
                minion_id,
                vm_ret["status"],
                vm_ret["error"],
            )
            return []
        else:
            vm_results.extend(vm_ret["dict"]["results"])
        # Check if we need to paginate and fetch the next result list
        if vm_ret["dict"]["next"]:
            vm_ret = salt.utils.http.query(
                vm_ret["dict"]["next"], header_dict=headers, decode=True
            )
        else:
            break

    # Set the node type
    vm_count = 0
    for vm in vm_results:
        vm_results[vm_count]["node_type"] = "virtual-machine"
        vm_count += 1

    # Return the results
    return vm_results


def _get_interfaces(
    api_url, minion_id, node_id, node_type, headers, api_query_result_limit
):
    log.debug(
        'Retrieving interfaces for "%s"',
        node_id,
    )
    interfaces_results = []
    if node_type == "device":
        app_name = "dcim"
        node_param = "device_id"
    elif node_type == "virtual-machine":
        app_name = "virtualization"
        node_param = "virtual_machine_id"
    interfaces_url = "{api_url}/{app}/{endpoint}".format(
        api_url=api_url, app=app_name, endpoint="interfaces"
    )
    params = {node_param: node_id}
    if api_query_result_limit:
        params["limit"] = api_query_result_limit
    interfaces_ret = salt.utils.http.query(
        interfaces_url,
        params=params,
        header_dict=headers,
        decode=True,
    )
    while True:
        # Check status code for API call
        if "error" in interfaces_ret:
            log.error(
                'Unable to retrieve interfaces for "%s" (Type %s, ID %d), status code: %d, error %s',
                minion_id,
                node_type,
                node_id,
                interfaces_ret["status"],
                interfaces_ret["error"],
            )
            return []
        else:
            interfaces_results.extend(interfaces_ret["dict"]["results"])
        # Check if we need to paginate and fetch the next result list
        if interfaces_ret["dict"]["next"]:
            interfaces_ret = salt.utils.http.query(
                interfaces_ret["dict"]["next"], header_dict=headers, decode=True
            )
        else:
            break

    # Clean up duplicate data in the dictionary
    interface_count = 0
    for interface in interfaces_results:
        if node_type == "device":
            del interfaces_results[interface_count]["device"]
        elif node_type == "virtual-machine":
            del interfaces_results[interface_count]["virtual_machine"]
        interface_count += 1

    # Return the results
    return interfaces_results


def _get_interface_ips(
    api_url, minion_id, node_id, node_type, headers, api_query_result_limit
):
    # We get all the IP addresses for the node at once instead of
    # having to make a separate call for each interface
    log.debug(
        'Retrieving IP addresses for "%s"',
        node_id,
    )
    interface_ips_results = []
    if node_type == "device":
        app_name = "dcim"
        node_param = "device_id"
    elif node_type == "virtual-machine":
        app_name = "virtualization"
        node_param = "virtual_machine_id"
    interface_ips_url = "{api_url}/{app}/{endpoint}".format(
        api_url=api_url, app="ipam", endpoint="ip-addresses"
    )
    params = {node_param: node_id}
    if api_query_result_limit:
        params["limit"] = api_query_result_limit
    interface_ips_ret = salt.utils.http.query(
        interface_ips_url,
        params=params,
        header_dict=headers,
        decode=True,
    )

    while True:
        # Check status code for API call
        if "error" in interface_ips_ret:
            log.error(
                'Unable to retrieve interface IP addresses for "%s" (Type %s, ID %d), status code: %d, error %s',
                minion_id,
                node_type,
                node_id,
                interface_ips_ret["status"],
                interface_ips_ret["error"],
            )
            return []
        else:
            interface_ips_results.extend(interface_ips_ret["dict"]["results"])
        # Check if we need to paginate and fetch the next result list
        if interface_ips_ret["dict"]["next"]:
            interface_ips_ret = salt.utils.http.query(
                interface_ips_ret["dict"]["next"], header_dict=headers, decode=True
            )
        else:
            break

    # Return the results
    return interface_ips_results


def _associate_ips_to_interfaces(interfaces_list, interface_ips_list):
    interface_count = 0
    for interface in interfaces_list:
        if len(interface_ips_list) > 0:
            interfaces_list[interface_count]["ip_addresses"] = []
            for ip in interface_ips_list:
                if (
                    "assigned_object_id" in ip
                    and ip["assigned_object_id"] == interface["id"]
                ):
                    del ip["assigned_object_type"]
                    del ip["assigned_object_id"]
                    del ip["assigned_object"]
                    interfaces_list[interface_count]["ip_addresses"].append(ip)
        interface_count += 1
    return interfaces_list


def _get_site_details(api_url, minion_id, site_name, site_id, headers):
    log.debug(
        'Retrieving site details for "%s" - site %s (ID %d)',
        minion_id,
        site_name,
        site_id,
    )
    site_url = "{api_url}/{app}/{endpoint}/{site_id}/".format(
        api_url=api_url, app="dcim", endpoint="sites", site_id=site_id
    )
    site_details_ret = salt.utils.http.query(site_url, header_dict=headers, decode=True)
    if "error" in site_details_ret:
        log.error(
            "Unable to retrieve site details for %s (ID %d), status code: %d, error %s",
            site_name,
            site_id,
            site_details_ret["status"],
            site_details_ret["error"],
        )
        return {}
    else:
        # Return the results
        return site_details_ret["dict"]


def _get_site_prefixes(
    api_url, minion_id, site_name, site_id, headers, api_query_result_limit
):
    log.debug(
        'Retrieving site prefixes for "%s" - site %s (ID %d)',
        minion_id,
        site_name,
        site_id,
    )
    site_prefixes_results = []
    prefixes_url = "{api_url}/{app}/{endpoint}".format(
        api_url=api_url, app="ipam", endpoint="prefixes"
    )
    params = {"site_id": site_id}
    if api_query_result_limit:
        params["limit"] = api_query_result_limit
    site_prefixes_ret = salt.utils.http.query(
        prefixes_url, params=params, header_dict=headers, decode=True
    )

    while True:
        # Check status code for API call
        if "error" in site_prefixes_ret:
            log.error(
                "Unable to retrieve site prefixes for %s (ID %d), status code: %d, error %s",
                site_name,
                site_id,
                site_prefixes_ret["status"],
                site_prefixes_ret["error"],
            )
            return []
        else:
            site_prefixes_results.extend(site_prefixes_ret["dict"]["results"])
        # Check if we need to paginate and fetch the next result list
        if site_prefixes_ret["dict"]["next"]:
            site_prefixes_ret = salt.utils.http.query(
                site_prefixes_ret["dict"]["next"], header_dict=headers, decode=True
            )
        else:
            break

    # Clean up duplicate data in the dictionary
    prefix_count = 0
    for prefix in site_prefixes_results:
        del site_prefixes_results[prefix_count]["site"]
        prefix_count += 1

    # Return the results
    return site_prefixes_results


def _get_proxy_details(api_url, minion_id, primary_ip, platform_id, headers):
    log.debug(
        'Retrieving proxy details for "%s"',
        minion_id,
    )
    platform_url = "{api_url}/{app}/{endpoint}/{id}/".format(
        api_url=api_url, app="dcim", endpoint="platforms", id=platform_id
    )
    platform_ret = salt.utils.http.query(platform_url, header_dict=headers, decode=True)
    # Check status code for API call
    if "error" in platform_ret:
        log.error(
            "Unable to proxy details for %s, status code: %d, error %s",
            minion_id,
            platform_ret["status"],
            platform_ret["error"],
        )
    else:
        # Assign results from API call to "proxy" key if the platform has a
        # napalm_driver defined.
        napalm_driver = platform_ret["dict"].get("napalm_driver")
        if napalm_driver:
            proxy = {
                "host": str(ipaddress.ip_interface(primary_ip).ip),
                "driver": napalm_driver,
                "proxytype": "napalm",
            }
            return proxy


def ext_pillar(minion_id, pillar, *args, **kwargs):
    """
    Query NetBox API for minion data
    """
    if minion_id == "*":
        log.info("There's no data to collect from NetBox for the Master")
        return {}
    # Pull settings from kwargs
    api_url = kwargs["api_url"].rstrip("/")
    api_token = kwargs.get("api_token")
    devices = kwargs.get("devices", True)
    virtual_machines = kwargs.get("virtual_machines", False)
    interfaces = kwargs.get("interfaces", False)
    interface_ips = kwargs.get("interface_ips", False)
    site_details = kwargs.get("site_details", True)
    site_prefixes = kwargs.get("site_prefixes", True)
    proxy_username = kwargs.get("proxy_username", None)
    proxy_return = kwargs.get("proxy_return", True)
    api_query_result_limit = kwargs.get("api_query_result_limit")

    ret = {}

    # Check that we have a valid API URL:
    if not salt.utils.url.validate(api_url, ["http", "https"]):
        log.error(
            'Provided URL for api_url "%s" is malformed or is not an http/https URL',
            api_url,
        )
        return ret

    # Check that the user has enabled at least one of the node options
    if not devices and not virtual_machines:
        log.error("At least one of devices or virtual_machines must be True")
        return ret

    # Check that the user has enabled interfaces if they've enabled interface_ips
    if interface_ips and not interfaces:
        log.error("The value for interfaces must be True if interface_ips is True")
        return ret

        # Check that the user has enabled interfaces if they've enabled interface_ips
    if api_query_result_limit and int(api_query_result_limit) <= 0:
        log.error(
            "The value for api_query_result_limit must be a postive integer if set"
        )
        return ret

    # Fetch device from API
    headers = {}
    if api_token:
        headers = {"Authorization": "Token {}".format(api_token)}
    else:
        log.error("The value for api_token is not set")
        return ret
    nodes = []
    if devices:
        nodes.extend(_get_devices(api_url, minion_id, headers, api_query_result_limit))
    if virtual_machines:
        nodes.extend(
            _get_virtual_machines(api_url, minion_id, headers, api_query_result_limit)
        )
    if len(nodes) == 1:
        # Return the 0th (and only) item in the list
        ret["netbox"] = nodes[0]
    elif len(nodes) > 1:
        log.error('More than one node found for "%s"', minion_id)
        return ret
    else:
        log.error('Unable to pull NetBox data for "%s"', minion_id)
        return ret
    node_id = ret["netbox"]["id"]
    node_type = ret["netbox"]["node_type"]
    if interfaces:
        interfaces_list = _get_interfaces(
            api_url, minion_id, node_id, node_type, headers, api_query_result_limit
        )
        if len(interfaces_list) > 0 and interface_ips:
            interface_ips_list = _get_interface_ips(
                api_url, minion_id, node_id, node_type, headers, api_query_result_limit
            )
            ret["netbox"]["interfaces"] = _associate_ips_to_interfaces(
                interfaces_list, interface_ips_list
            )
    site_id = ret["netbox"]["site"]["id"]
    site_name = ret["netbox"]["site"]["name"]
    if site_details:
        ret["netbox"]["site"] = _get_site_details(
            api_url, minion_id, site_name, site_id, headers
        )
    if site_prefixes:
        ret["netbox"]["site"]["prefixes"] = _get_site_prefixes(
            api_url, minion_id, site_name, site_id, headers, api_query_result_limit
        )
    if proxy_return:
        platform_id = ret["netbox"]["platform"]["id"]
        primary_ip = ret["netbox"]["primary_ip"]["address"]
        proxy = _get_proxy_details(api_url, minion_id, primary_ip, platform_id, headers)
        if proxy:
            ret["proxy"] = proxy
            if proxy_username:
                ret["proxy"]["username"] = proxy_username

    return ret



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import salt.modules.lvs as lvs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class LvsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.lvs
    """

    def setup_loader_modules(self):
        return {lvs: {}}

    def test_add_service(self):
        """
        Test for Add a virtual service.
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.object(lvs, "_build_cmd", return_value="B"):
                with patch.dict(
                    lvs.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={"retcode": "ret", "stderr": "stderr"}
                        )
                    },
                ):
                    self.assertEqual(lvs.add_service(), "stderr")

    def test_edit_service(self):
        """
        Test for Edit the virtual service.
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.object(lvs, "_build_cmd", return_value="B"):
                with patch.dict(
                    lvs.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={"retcode": "ret", "stderr": "stderr"}
                        )
                    },
                ):
                    self.assertEqual(lvs.edit_service(), "stderr")

    def test_delete_service(self):
        """
        Test for Delete the virtual service.
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.object(lvs, "_build_cmd", return_value="B"):
                with patch.dict(
                    lvs.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={"retcode": "ret", "stderr": "stderr"}
                        )
                    },
                ):
                    self.assertEqual(lvs.delete_service(), "stderr")

    def test_add_server(self):
        """
        Test for Add a real server to a virtual service.
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.object(lvs, "_build_cmd", return_value="B"):
                with patch.dict(
                    lvs.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={"retcode": "ret", "stderr": "stderr"}
                        )
                    },
                ):
                    self.assertEqual(lvs.add_server(), "stderr")

    def test_edit_server(self):
        """
        Test for Edit a real server to a virtual service.
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.object(lvs, "_build_cmd", return_value="B"):
                with patch.dict(
                    lvs.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={"retcode": "ret", "stderr": "stderr"}
                        )
                    },
                ):
                    self.assertEqual(lvs.edit_server(), "stderr")

    def test_delete_server(self):
        """
        Test for Delete the realserver from the virtual service.
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.object(lvs, "_build_cmd", return_value="B"):
                with patch.dict(
                    lvs.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={"retcode": "ret", "stderr": "stderr"}
                        )
                    },
                ):
                    self.assertEqual(lvs.delete_server(), "stderr")

    def test_clear(self):
        """
        Test for Clear the virtual server table
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                self.assertEqual(lvs.clear(), "stderr")

    def test_get_rules(self):
        """
        Test for Get the virtual server rules
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.dict(lvs.__salt__, {"cmd.run": MagicMock(return_value="A")}):
                self.assertEqual(lvs.get_rules(), "A")

    def test_list_(self):
        """
        Test for List the virtual server table
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.object(lvs, "_build_cmd", return_value="B"):
                with patch.dict(
                    lvs.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={"retcode": "ret", "stderr": "stderr"}
                        )
                    },
                ):
                    self.assertEqual(lvs.list_("p", "s"), "stderr")

    def test_zero(self):
        """
        Test for Zero the packet, byte and rate counters in a
         service or all services.
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.object(lvs, "_build_cmd", return_value="B"):
                with patch.dict(
                    lvs.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={"retcode": "ret", "stderr": "stderr"}
                        )
                    },
                ):
                    self.assertEqual(lvs.zero("p", "s"), "stderr")

    def test_check_service(self):
        """
        Test for Check the virtual service exists.
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.object(lvs, "_build_cmd", return_value="B"):
                with patch.dict(
                    lvs.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={"retcode": "ret", "stderr": "stderr"}
                        )
                    },
                ):
                    with patch.object(lvs, "get_rules", return_value="C"):
                        self.assertEqual(
                            lvs.check_service("p", "s"), "Error: service not exists"
                        )

    def test_check_server(self):
        """
        Test for Check the real server exists in the specified service.
        """
        with patch.object(lvs, "__detect_os", return_value="C"):
            with patch.object(lvs, "_build_cmd", return_value="B"):
                with patch.dict(
                    lvs.__salt__,
                    {
                        "cmd.run_all": MagicMock(
                            return_value={"retcode": "ret", "stderr": "stderr"}
                        )
                    },
                ):
                    with patch.object(lvs, "get_rules", return_value="C"):
                        self.assertEqual(
                            lvs.check_server("p", "s"), "Error: server not exists"
                        )



"""
unittests for highstate outputter
"""


import salt.output.highstate as highstate
import salt.utils.stringutils
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase


class JsonTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.output.highstate
    """

    def setup_loader_modules(self):
        return {
            highstate: {
                "__opts__": {
                    "extension_modules": "",
                    "optimization_order": [0, 1, 2],
                    "color": False,
                }
            }
        }

    def setUp(self):
        self.data = {
            "data": {
                "master": {
                    "salt_|-call_sleep_state_|-call_sleep_state_|-state": {
                        "__id__": "call_sleep_state",
                        "__jid__": "20170418153529810135",
                        "__run_num__": 0,
                        "__sls__": "orch.simple",
                        "changes": {
                            "out": "highstate",
                            "ret": {
                                "minion": {
                                    "module_|-simple-ping_|-test.ping_|-run": {
                                        "__id__": "simple-ping",
                                        "__run_num__": 0,
                                        "__sls__": "simple-ping",
                                        "changes": {"ret": True},
                                        "comment": "Module function test.ping executed",
                                        "duration": 56.179,
                                        "name": "test.ping",
                                        "result": True,
                                        "start_time": "15:35:31.282099",
                                    }
                                },
                                "sub_minion": {
                                    "module_|-simple-ping_|-test.ping_|-run": {
                                        "__id__": "simple-ping",
                                        "__run_num__": 0,
                                        "__sls__": "simple-ping",
                                        "changes": {"ret": True},
                                        "comment": "Module function test.ping executed",
                                        "duration": 54.103,
                                        "name": "test.ping",
                                        "result": True,
                                        "start_time": "15:35:31.005606",
                                    }
                                },
                            },
                        },
                        "comment": (
                            "States ran successfully. Updating sub_minion, minion."
                        ),
                        "duration": 1638.047,
                        "name": "call_sleep_state",
                        "result": True,
                        "start_time": "15:35:29.762657",
                    },
                    "salt_|-cmd_run_example_|-cmd.run_|-function": {
                        "__id__": "cmd_run_example",
                        "__jid__": "20200411195112288850",
                        "__run_num__": 1,
                        "__sls__": "orch.simple",
                        "changes": {
                            "out": "highstate",
                            "ret": {"minion": "file1\nfile2\nfile3"},
                        },
                        "comment": (
                            "Function ran successfully. Function cmd.run ran on minion."
                        ),
                        "duration": 412.397,
                        "name": "cmd.run",
                        "result": True,
                        "start_time": "21:51:12.185868",
                    },
                }
            },
            "outputter": "highstate",
            "retcode": 0,
        }
        self.addCleanup(delattr, self, "data")

    def test_default_output(self):
        ret = highstate.output(self.data)
        self.assertIn("Succeeded: 1 (changed=1)", ret)
        self.assertIn("Failed:    0", ret)
        self.assertIn("Total states run:     1", ret)
        self.assertIn("                  file2", ret)

    def test_output_comment_is_not_unicode(self):
        entry = None
        for key in (
            "data",
            "master",
            "salt_|-call_sleep_state_|-call_sleep_state_|-state",
            "changes",
            "ret",
            "minion",
            "module_|-simple-ping_|-test.ping_|-run",
        ):
            if entry is None:
                entry = self.data[key]
                continue
            entry = entry[key]
        entry["comment"] = salt.utils.stringutils.to_bytes(entry["comment"])
        ret = highstate.output(self.data)
        self.assertIn("Succeeded: 1 (changed=1)", ret)
        self.assertIn("Failed:    0", ret)
        self.assertIn("Total states run:     1", ret)
        self.assertIn("                  file2", ret)


# this should all pass the above tests
class JsonNestedTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for nested salt.output.highstate (ie orchestrations calling other orchs)
    """

    def setup_loader_modules(self):
        return {
            highstate: {
                "__opts__": {
                    "extension_modules": "",
                    "color": False,
                    "state_output_profile": True,
                    "optimization_order": [0, 1, 2],
                }
            }
        }

    def setUp(self):
        self.data = {
            "outputter": "highstate",
            "data": {
                "local_master": {
                    "salt_|-nested_|-state.orchestrate_|-runner": {
                        "comment": "Runner function 'state.orchestrate' executed.",
                        "name": "state.orchestrate",
                        "__orchestration__": True,
                        "start_time": "09:22:53.158742",
                        "result": True,
                        "duration": 980.694,
                        "__run_num__": 0,
                        "__jid__": "20180326092253538853",
                        "__sls__": "orch.test.nested",
                        "changes": {
                            "return": {
                                "outputter": "highstate",
                                "data": {
                                    "local_master": {
                                        "test_|-always-passes-with-changes_|-oinaosf_|-succeed_with_changes": {
                                            "comment": "Success!",
                                            "name": "oinaosf",
                                            "start_time": "09:22:54.128415",
                                            "result": True,
                                            "duration": 0.437,
                                            "__run_num__": 0,
                                            "__sls__": "orch.test.changes",
                                            "changes": {
                                                "testing": {
                                                    "new": (
                                                        "Something pretended to change"
                                                    ),
                                                    "old": "Unchanged",
                                                }
                                            },
                                            "__id__": "always-passes-with-changes",
                                        },
                                        "test_|-always-passes_|-fasdfasddfasdfoo_|-succeed_without_changes": {
                                            "comment": "Success!",
                                            "name": "fasdfasddfasdfoo",
                                            "start_time": "09:22:54.128986",
                                            "result": True,
                                            "duration": 0.25,
                                            "__run_num__": 1,
                                            "__sls__": "orch.test.changes",
                                            "changes": {},
                                            "__id__": "always-passes",
                                        },
                                    }
                                },
                                "retcode": 0,
                            }
                        },
                        "__id__": "nested",
                    }
                }
            },
            "retcode": 0,
        }

        self.addCleanup(delattr, self, "data")

    def test_nested_output(self):
        ret = highstate.output(self.data)
        self.assertIn("Succeeded: 1 (changed=1)", ret)
        self.assertIn("Failed:    0", ret)
        self.assertIn("Total states run:     1", ret)

        # the whitespace is relevant in this case, it is testing that it is nested
        self.assertIn("                        ID: always-passes-with-changes", ret)
        self.assertIn("                   Started: 09:22:54.128415", ret)
        self.assertIn("              Succeeded: 2 (changed=1)", ret)
        self.assertIn("              Failed:    0", ret)
        self.assertIn("              Total states run:     2", ret)



"""
This is the default compound matcher function.
"""

import logging

import salt.loader
import salt.utils.minions

HAS_RANGE = False
try:
    import seco.range  # pylint: disable=unused-import

    HAS_RANGE = True
except ImportError:
    pass

log = logging.getLogger(__name__)


def match(tgt, opts=None, minion_id=None):
    """
    Runs the compound target check
    """
    if not opts:
        opts = __opts__
    nodegroups = opts.get("nodegroups", {})
    matchers = salt.loader.matchers(opts)
    if not minion_id:
        minion_id = opts.get("id")

    if not isinstance(tgt, str) and not isinstance(tgt, (list, tuple)):
        log.error("Compound target received that is neither string, list nor tuple")
        return False
    log.debug("compound_match: %s ? %s", minion_id, tgt)
    ref = {
        "G": "grain",
        "P": "grain_pcre",
        "I": "pillar",
        "J": "pillar_pcre",
        "L": "list",
        "N": None,  # Nodegroups should already be expanded
        "S": "ipcidr",
        "E": "pcre",
    }
    if HAS_RANGE:
        ref["R"] = "range"

    results = []
    opers = ["and", "or", "not", "(", ")"]

    if isinstance(tgt, str):
        words = tgt.split()
    else:
        # we make a shallow copy in order to not affect the passed in arg
        words = tgt[:]

    while words:
        word = words.pop(0)
        target_info = salt.utils.minions.parse_target(word)

        # Easy check first
        if word in opers:
            if results:
                if results[-1] == "(" and word in ("and", "or"):
                    log.error('Invalid beginning operator after "(": %s', word)
                    return False
                if word == "not":
                    if not results[-1] in ("and", "or", "("):
                        results.append("and")
                results.append(word)
            else:
                # seq start with binary oper, fail
                if word not in ["(", "not"]:
                    log.error("Invalid beginning operator: %s", word)
                    return False
                results.append(word)

        elif target_info and target_info["engine"]:
            if "N" == target_info["engine"]:
                # if we encounter a node group, just evaluate it in-place
                decomposed = salt.utils.minions.nodegroup_comp(
                    target_info["pattern"], nodegroups
                )
                if decomposed:
                    words = decomposed + words
                continue

            engine = ref.get(target_info["engine"])
            if not engine:
                # If an unknown engine is called at any time, fail out
                log.error(
                    'Unrecognized target engine "%s" for target expression "%s"',
                    target_info["engine"],
                    word,
                )
                return False

            engine_args = [target_info["pattern"]]
            engine_kwargs = {"opts": opts, "minion_id": minion_id}
            if target_info["delimiter"]:
                engine_kwargs["delimiter"] = target_info["delimiter"]

            results.append(
                str(
                    matchers["{}_match.match".format(engine)](
                        *engine_args, **engine_kwargs
                    )
                )
            )

        else:
            # The match is not explicitly defined, evaluate it as a glob
            results.append(str(matchers["glob_match.match"](word, opts, minion_id)))

    results = " ".join(results)
    log.debug('compound_match %s ? "%s" => "%s"', minion_id, tgt, results)
    try:
        return eval(results)  # pylint: disable=W0123
    except Exception:  # pylint: disable=broad-except
        log.error("Invalid compound target: %s for results: %s", tgt, results)
    return False



"""
    salt._logging.impl
    ~~~~~~~~~~~~~~~~~~

    Salt's logging implementation classes/functionality
"""
import logging
import re
import sys
import types

# Let's define these custom logging levels before importing the salt._logging.mixins
# since they will be used there
PROFILE = logging.PROFILE = 15
TRACE = logging.TRACE = 5
GARBAGE = logging.GARBAGE = 1
QUIET = logging.QUIET = 1000

from salt._logging.handlers import StreamHandler  # isort:skip

# from salt._logging.handlers import SysLogHandler  # isort:skip
# from salt._logging.handlers import RotatingFileHandler  # isort:skip
# from salt._logging.handlers import WatchedFileHandler  # isort:skip
from salt._logging.handlers import TemporaryLoggingHandler  # isort:skip
from salt._logging.mixins import LoggingMixinMeta  # isort:skip
from salt._logging.mixins import NewStyleClassMixin  # isort:skip
from salt.exceptions import LoggingRuntimeError  # isort:skip
from salt.utils.ctx import RequestContext  # isort:skip
from salt.utils.textformat import TextFormat  # isort:skip

LOG_LEVELS = {
    "all": logging.NOTSET,
    "debug": logging.DEBUG,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    "garbage": GARBAGE,
    "info": logging.INFO,
    "profile": PROFILE,
    "quiet": QUIET,
    "trace": TRACE,
    "warning": logging.WARNING,
}

LOG_VALUES_TO_LEVELS = {v: k for (k, v) in LOG_LEVELS.items()}

LOG_COLORS = {
    "levels": {
        "QUIET": TextFormat("reset"),
        "CRITICAL": TextFormat("bold", "red"),
        "ERROR": TextFormat("bold", "red"),
        "WARNING": TextFormat("bold", "yellow"),
        "INFO": TextFormat("bold", "green"),
        "PROFILE": TextFormat("bold", "cyan"),
        "DEBUG": TextFormat("bold", "cyan"),
        "TRACE": TextFormat("bold", "magenta"),
        "GARBAGE": TextFormat("bold", "blue"),
        "NOTSET": TextFormat("reset"),
        "SUBDEBUG": TextFormat(
            "bold", "cyan"
        ),  # used by multiprocessing.log_to_stderr()
        "SUBWARNING": TextFormat(
            "bold", "yellow"
        ),  # used by multiprocessing.log_to_stderr()
    },
    "msgs": {
        "QUIET": TextFormat("reset"),
        "CRITICAL": TextFormat("bold", "red"),
        "ERROR": TextFormat("red"),
        "WARNING": TextFormat("yellow"),
        "INFO": TextFormat("green"),
        "PROFILE": TextFormat("bold", "cyan"),
        "DEBUG": TextFormat("cyan"),
        "TRACE": TextFormat("magenta"),
        "GARBAGE": TextFormat("blue"),
        "NOTSET": TextFormat("reset"),
        "SUBDEBUG": TextFormat(
            "bold", "cyan"
        ),  # used by multiprocessing.log_to_stderr()
        "SUBWARNING": TextFormat(
            "bold", "yellow"
        ),  # used by multiprocessing.log_to_stderr()
    },
    "name": TextFormat("bold", "green"),
    "process": TextFormat("bold", "blue"),
}

# Make a list of log level names sorted by log level
SORTED_LEVEL_NAMES = [l[0] for l in sorted(LOG_LEVELS.items(), key=lambda x: x[1])]

MODNAME_PATTERN = re.compile(r"(?P<name>%%\(name\)(?:\-(?P<digits>[\d]+))?s)")


# ----- REMOVE ME ON REFACTOR COMPLETE ------------------------------------------------------------------------------>
class __NullLoggingHandler(TemporaryLoggingHandler):
    """
    This class exists just to better identify which temporary logging
    handler is being used for what.
    """


class __StoreLoggingHandler(TemporaryLoggingHandler):
    """
    This class exists just to better identify which temporary logging
    handler is being used for what.
    """


# Store a reference to the temporary queue logging handler
LOGGING_NULL_HANDLER = __NullLoggingHandler(logging.WARNING)

# Store a reference to the temporary console logger
LOGGING_TEMP_HANDLER = StreamHandler(sys.stderr)

# Store a reference to the "storing" logging handler
LOGGING_STORE_HANDLER = __StoreLoggingHandler()
# <---- REMOVE ME ON REFACTOR COMPLETE -------------------------------------------------------------------------------


class SaltLogRecord(logging.LogRecord):
    def __init__(self, *args, **kwargs):
        logging.LogRecord.__init__(self, *args, **kwargs)
        self.bracketname = "[{:<17}]".format(str(self.name))
        self.bracketlevel = "[{:<8}]".format(str(self.levelname))
        self.bracketprocess = "[{:>5}]".format(str(self.process))


class SaltColorLogRecord(SaltLogRecord):
    def __init__(self, *args, **kwargs):
        SaltLogRecord.__init__(self, *args, **kwargs)

        reset = TextFormat("reset")
        clevel = LOG_COLORS["levels"].get(self.levelname, reset)
        cmsg = LOG_COLORS["msgs"].get(self.levelname, reset)

        self.colorname = "{}[{:<17}]{}".format(
            LOG_COLORS["name"], str(self.name), reset
        )
        self.colorlevel = "{}[{:<8}]{}".format(clevel, str(self.levelname), reset)
        self.colorprocess = "{}[{:>5}]{}".format(
            LOG_COLORS["process"], str(self.process), reset
        )
        self.colormsg = "{}{}{}".format(cmsg, self.getMessage(), reset)


def get_log_record_factory():
    """
    Get the logging  log record factory
    """
    try:
        return get_log_record_factory.__factory__
    except AttributeError:
        return


def set_log_record_factory(factory):
    """
    Set the logging  log record factory
    """
    get_log_record_factory.__factory__ = factory
    logging.setLogRecordFactory(factory)


set_log_record_factory(SaltLogRecord)


# Store an instance of the current logging logger class
LOGGING_LOGGER_CLASS = logging.getLoggerClass()


class SaltLoggingClass(
    LOGGING_LOGGER_CLASS, NewStyleClassMixin, metaclass=LoggingMixinMeta
):
    def __new__(cls, *args):
        """
        We override `__new__` in our logging logger class in order to provide
        some additional features like expand the module name padding if length
        is being used, and also some Unicode fixes.

        This code overhead will only be executed when the class is
        instantiated, i.e.:

            logging.getLogger(__name__)

        """
        instance = super().__new__(cls)

        try:
            max_logger_length = len(
                max(list(logging.Logger.manager.loggerDict), key=len)
            )
            for handler in logging.root.handlers:
                if handler in (
                    LOGGING_NULL_HANDLER,
                    LOGGING_STORE_HANDLER,
                    LOGGING_TEMP_HANDLER,
                ):
                    continue

                formatter = handler.formatter
                if not formatter:
                    continue

                if not handler.lock:
                    handler.createLock()
                handler.acquire()

                fmt = formatter._fmt.replace("%", "%%")

                match = MODNAME_PATTERN.search(fmt)
                if not match:
                    # Not matched. Release handler and return.
                    handler.release()
                    return instance

                if "digits" not in match.groupdict():
                    # No digits group. Release handler and return.
                    handler.release()
                    return instance

                digits = match.group("digits")
                if not digits or not (digits and digits.isdigit()):
                    # No valid digits. Release handler and return.
                    handler.release()
                    return instance

                if int(digits) < max_logger_length:
                    # Formatter digits value is lower than current max, update.
                    fmt = fmt.replace(match.group("name"), "%%(name)-%ds")
                    formatter = logging.Formatter(
                        fmt % max_logger_length, datefmt=formatter.datefmt
                    )
                    handler.setFormatter(formatter)
                handler.release()
        except ValueError:
            # There are no registered loggers yet
            pass
        return instance

    def _log(
        self,
        level,
        msg,
        args,
        exc_info=None,
        extra=None,  # pylint: disable=arguments-differ
        stack_info=False,
        stacklevel=1,
        exc_info_on_loglevel=None,
    ):
        if extra is None:
            extra = {}

        # pylint: disable=no-member
        current_jid = RequestContext.current.get("data", {}).get("jid", None)
        log_fmt_jid = RequestContext.current.get("opts", {}).get("log_fmt_jid", None)
        # pylint: enable=no-member

        if current_jid is not None:
            extra["jid"] = current_jid

        if log_fmt_jid is not None:
            extra["log_fmt_jid"] = log_fmt_jid

        # If both exc_info and exc_info_on_loglevel are both passed, let's fail
        if exc_info and exc_info_on_loglevel:
            raise LoggingRuntimeError(
                "Only one of 'exc_info' and 'exc_info_on_loglevel' is permitted"
            )
        if exc_info_on_loglevel is not None:
            if isinstance(exc_info_on_loglevel, str):
                exc_info_on_loglevel = LOG_LEVELS.get(
                    exc_info_on_loglevel, logging.ERROR
                )
            elif not isinstance(exc_info_on_loglevel, int):
                raise RuntimeError(
                    "The value of 'exc_info_on_loglevel' needs to be a "
                    "logging level or a logging level name, not '{}'".format(
                        exc_info_on_loglevel
                    )
                )
        if extra is None:
            extra = {"exc_info_on_loglevel": exc_info_on_loglevel}
        else:
            extra["exc_info_on_loglevel"] = exc_info_on_loglevel

        if sys.version_info < (3,):
            LOGGING_LOGGER_CLASS._log(
                self, level, msg, args, exc_info=exc_info, extra=extra
            )
        elif sys.version_info < (3, 8):
            LOGGING_LOGGER_CLASS._log(
                self,
                level,
                msg,
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
            )
        else:
            LOGGING_LOGGER_CLASS._log(
                self,
                level,
                msg,
                args,
                exc_info=exc_info,
                extra=extra,
                stack_info=stack_info,
                stacklevel=stacklevel,
            )

    def makeRecord(
        self,
        name,
        level,
        fn,
        lno,
        msg,
        args,
        exc_info,
        func=None,
        extra=None,
        sinfo=None,
    ):
        # Let's remove exc_info_on_loglevel from extra
        exc_info_on_loglevel = extra.pop("exc_info_on_loglevel")

        jid = extra.pop("jid", "")
        if jid:
            log_fmt_jid = extra.pop("log_fmt_jid")
            jid = log_fmt_jid % {"jid": jid}

        if not extra:
            # If nothing else is in extra, make it None
            extra = None

        # Let's try to make every logging message unicode
        try:
            salt_system_encoding = __salt_system_encoding__
            if salt_system_encoding == "ascii":
                # Encoding detection most likely failed, let's use the utf-8
                # value which we defaulted before __salt_system_encoding__ was
                # implemented
                salt_system_encoding = "utf-8"
        except NameError:
            salt_system_encoding = "utf-8"

        if isinstance(msg, bytes):
            try:
                _msg = msg.decode(salt_system_encoding, "replace")
            except UnicodeDecodeError:
                _msg = msg.decode(salt_system_encoding, "ignore")
        else:
            _msg = msg

        _args = []
        for item in args:
            if isinstance(item, bytes):
                try:
                    _args.append(item.decode(salt_system_encoding, "replace"))
                except UnicodeDecodeError:
                    _args.append(item.decode(salt_system_encoding, "ignore"))
            else:
                _args.append(item)
        _args = tuple(_args)

        logrecord = LOGGING_LOGGER_CLASS.makeRecord(
            self, name, level, fn, lno, _msg, _args, exc_info, func, sinfo
        )

        if exc_info_on_loglevel is not None:
            # Let's add some custom attributes to the LogRecord class in order
            # to include the exc_info on a per handler basis. This will allow
            # showing tracebacks on logfiles but not on console if the logfile
            # handler is enabled for the log level "exc_info_on_loglevel" and
            # console handler is not.
            logrecord.exc_info_on_loglevel_instance = sys.exc_info()
            logrecord.exc_info_on_loglevel_formatted = None

        logrecord.exc_info_on_loglevel = exc_info_on_loglevel
        logrecord.jid = jid
        return logrecord


# Override the python's logging logger class as soon as this module is imported
if logging.getLoggerClass() is not SaltLoggingClass:

    # Import pip._internal which itself will install it's own custom logging handler
    # we want to override that handler with ours
    try:
        import pip._internal.utils._log as pip_log_module  # pylint: disable=no-name-in-module,import-error
    except ImportError:
        pip_log_module = None

    logging.setLoggerClass(SaltLoggingClass)
    logging.addLevelName(QUIET, "QUIET")
    logging.addLevelName(PROFILE, "PROFILE")
    logging.addLevelName(TRACE, "TRACE")
    logging.addLevelName(GARBAGE, "GARBAGE")
    if pip_log_module is not None:
        # Let's make newer versions of pip work by patching SaltLoggingClass to
        # add a verbose method which is what pip expects
        SaltLoggingClass.verbose = SaltLoggingClass.debug

    # ----- REMOVE ON REFACTORING COMPLETE -------------------------------------------------------------------------->
    if not logging.root.handlers:
        # No configuration to the logging system has been done so far.
        # Set the root logger at the lowest level possible
        logging.root.setLevel(GARBAGE)

        # Add a Null logging handler until logging is configured(will be
        # removed at a later stage) so we stop getting:
        #   No handlers could be found for logger 'foo'
        logging.root.addHandler(LOGGING_NULL_HANDLER)

    # Add the queue logging handler so we can later sync all message records
    # with the additional logging handlers
    logging.root.addHandler(LOGGING_STORE_HANDLER)
    # <---- REMOVE ON REFACTORING COMPLETE ---------------------------------------------------------------------------


# Now that we defined the default logging logger class, we can instantiate our logger
# DO NOT MOVE THIS
log = logging.getLogger(__name__)


def __get_exposed_module_attributes():
    """
    This function just ``dir()``'s this module and filters out any functions
    or variables which should not be available when wildcard importing it
    """
    exposed = []
    module = sys.modules[__name__]
    for name in dir(module):
        if name.startswith("_"):
            continue
        obj = getattr(module, name)
        if not isinstance(obj, types.FunctionType):
            if name.startswith(("LOG_", "SORTED_")):
                exposed.append(name)
            continue
        if obj.__module__ != __name__:
            continue
        exposed.append(name)
    return exposed


# Define what can be imported by wildcard imports
__all__ = __get_exposed_module_attributes()

# We're done with the function, nuke it
del __get_exposed_module_attributes



import os
import shutil
import textwrap

import pytest
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from tests.support.case import ModuleCase


@pytest.mark.windows_whitelisted
class PyDSLRendererIncludeTestCase(ModuleCase):
    def setUp(self):
        self.directory_created = False
        if salt.utils.platform.is_windows():
            if not os.path.isdir("\\tmp"):
                os.mkdir("\\tmp")
                self.directory_created = True

    def tearDown(self):
        if salt.utils.platform.is_windows():
            if self.directory_created:
                shutil.rmtree("\\tmp")

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_rendering_includes(self):
        """
        This test is currently hard-coded to /tmp to work-around a seeming
        inability to load custom modules inside the pydsl renderers. This
        is a FIXME.
        """
        self.run_function("state.sls", ["pydsl.aaa"])

        expected = textwrap.dedent(
            """\
            X1
            X2
            X3
            Y1 extended
            Y2 extended
            Y3
            hello red 1
            hello green 2
            hello blue 3
            """
        )

        # Windows adds `linefeed` in addition to `newline`. There's also an
        # unexplainable space before the `linefeed`...
        if salt.utils.platform.is_windows():
            expected = (
                "X1 \r\n"
                "X2 \r\n"
                "X3 \r\n"
                "Y1 extended \r\n"
                "Y2 extended \r\n"
                "Y3 \r\n"
                "hello red 1 \r\n"
                "hello green 2 \r\n"
                "hello blue 3 \r\n"
            )

        try:
            with salt.utils.files.fopen("/tmp/output", "r") as f:
                ret = salt.utils.stringutils.to_unicode(f.read())
        finally:
            os.remove("/tmp/output")

        self.assertEqual(sorted(ret), sorted(expected))



#!/usr/bin/env python
#
# Copyright 2011 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Implementation of platform-specific functionality.

For each function or class described in `tornado.platform.interface`,
the appropriate platform-specific implementation exists in this module.
Most code that needs access to this functionality should do e.g.::

    from salt.ext.tornado.platform.auto import set_close_exec
"""
# pylint: skip-file

from __future__ import absolute_import, division, print_function

import os

if 'APPENGINE_RUNTIME' in os.environ:
    from salt.ext.tornado.platform.common import Waker

    def set_close_exec(fd):
        pass
elif os.name == 'nt':
    from salt.ext.tornado.platform.common import Waker
    from salt.ext.tornado.platform.windows import set_close_exec
else:
    from salt.ext.tornado.platform.posix import set_close_exec, Waker

try:
    # monotime monkey-patches the time module to have a monotonic function
    # in versions of python before 3.3.
    import monotime
    # Silence pyflakes warning about this unused import
    monotime
except ImportError:
    pass
try:
    # monotonic can provide a monotonic function in versions of python before
    # 3.3, too.
    from monotonic import monotonic as monotonic_time
except ImportError:
    try:
        from time import monotonic as monotonic_time
    except ImportError:
        monotonic_time = None

__all__ = ['Waker', 'set_close_exec', 'monotonic_time']



"""
Version agnostic psutil hack to fully support both old (<2.0) and new (>=2.0)
psutil versions.

The old <1.0 psutil API is dropped in psutil 3.0

Should be removed once support for psutil <2.0 is dropped. (eg RHEL 6)

Built off of http://grodola.blogspot.com/2014/01/psutil-20-porting.html
"""


# No exception handling, as we want ImportError if psutil doesn't exist
import psutil  # pylint: disable=3rd-party-module-not-gated

if psutil.version_info >= (2, 0):
    from psutil import *  # pylint: disable=wildcard-import,unused-wildcard-import,3rd-party-module-not-gated
else:
    # Import hack to work around bugs in old psutil's
    # Psuedo "from psutil import *"
    _globals = globals()
    for attr in psutil.__all__:
        _temp = __import__("psutil", globals(), locals(), [attr], 0)
        try:
            _globals[attr] = getattr(_temp, attr)
        except AttributeError:
            pass

    # Import functions not in __all__
    # pylint: disable=unused-import,3rd-party-module-not-gated
    from psutil import disk_partitions
    from psutil import disk_usage

    # pylint: enable=unused-import,3rd-party-module-not-gated

    # Alias new module functions
    def boot_time():
        return psutil.BOOT_TIME

    def cpu_count():
        return psutil.NUM_CPUS

    # Alias renamed module functions
    pids = psutil.get_pid_list
    try:
        users = psutil.get_users
    except AttributeError:
        users = lambda: (_ for _ in ()).throw(
            NotImplementedError("Your psutil version is too old")
        )

    # Deprecated in 1.0.1, but not mentioned in blog post
    if psutil.version_info < (1, 0, 1):
        net_io_counters = psutil.network_io_counters()

    class Process(psutil.Process):  # pylint: disable=no-init
        # Reimplement overloaded getters/setters
        # pylint: disable=arguments-differ
        def cpu_affinity(self, *args, **kwargs):
            if args or kwargs:
                return self.set_cpu_affinity(*args, **kwargs)
            else:
                return self.get_cpu_affinity()

        def ionice(self, *args, **kwargs):
            if args or kwargs:
                return self.set_ionice(*args, **kwargs)
            else:
                return self.get_ionice()

        def nice(self, *args, **kwargs):
            if args or kwargs:
                return self.set_nice(*args, **kwargs)
            else:
                return self.get_nice()

        def rlimit(self, *args, **kwargs):
            """
            set_rlimit and get_limit were not introduced until psutil v1.1.0
            """
            if psutil.version_info >= (1, 1, 0):
                if args or kwargs:
                    return self.set_rlimit(*args, **kwargs)
                else:
                    return self.get_rlimit()
            else:
                pass

        # pylint: enable=arguments-differ

    # Alias renamed Process functions
    _PROCESS_FUNCTION_MAP = {
        "children": "get_children",
        "connections": "get_connections",
        "cpu_percent": "get_cpu_percent",
        "cpu_times": "get_cpu_times",
        "io_counters": "get_io_counters",
        "memory_info": "get_memory_info",
        "memory_info_ex": "get_ext_memory_info",
        "memory_maps": "get_memory_maps",
        "memory_percent": "get_memory_percent",
        "num_ctx_switches": "get_num_ctx_switches",
        "num_fds": "get_num_fds",
        "num_threads": "get_num_threads",
        "open_files": "get_open_files",
        "threads": "get_threads",
        "cwd": "getcwd",
    }

    for new, old in _PROCESS_FUNCTION_MAP.items():
        try:
            setattr(Process, new, psutil.Process.__dict__[old])
        except KeyError:
            pass



import salt.utils.platform
import salt.utils.win_network as win_network
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf


class PhysicalAddress:
    def __init__(self, address):
        self.address = address

    def ToString(self):
        return str(self.address)


class Interface:
    """
    Mocked interface object
    """

    def __init__(
        self,
        i_address="02D5F1DD31E0",
        i_description="Dell GigabitEthernet",
        i_id="{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
        i_name="Ethernet",
        i_receive_only=False,
        i_status=1,
        i_type=6,
    ):
        self.PhysicalAddress = PhysicalAddress(i_address)
        self.Description = i_description
        self.Id = i_id
        self.Name = i_name
        self.NetworkInterfaceType = i_type
        self.IsReceiveOnly = i_receive_only
        self.OperationalStatus = i_status

    def GetPhysicalAddress(self):
        return self.PhysicalAddress


@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
class WinNetworkTestCase(TestCase):
    def setUp(self):
        self.mock_ip_base = MagicMock(
            return_value={
                "dns_enabled": False,
                "dns_suffix": "",
                "dynamic_dns_enabled": False,
            }
        )
        self.mock_unicast = MagicMock(
            return_value={
                "ip_addresses": [
                    {
                        "address": "172.18.87.49",
                        "broadcast": "172.18.87.63",
                        "loopback": "127.0.0.1",
                        "netmask": "255.255.255.240",
                        "prefix_length": 28,
                        "prefix_origin": "Manual",
                        "suffix_origin": "Manual",
                    }
                ],
                "ipv6_addresses": [
                    {
                        "address": "fe80::e8a4:1224:5548:2b81",
                        "interface_index": 32,
                        "prefix_length": 64,
                        "prefix_origin": "WellKnown",
                        "suffix_origin": "Router",
                    }
                ],
            }
        )
        self.mock_gateway = MagicMock(
            return_value={
                "ip_gateways": ["192.168.0.1"],
                "ipv6_gateways": ["fe80::208:a2ff:fe0b:de70"],
            }
        )
        self.mock_dns = MagicMock(
            return_value={
                "ip_dns": ["10.4.0.1", "10.1.0.1", "8.8.8.8"],
                "ipv6_dns": ["2600:740a:1:304::1"],
            }
        )
        self.mock_multicast = MagicMock(
            return_value={
                "ip_multicast": [
                    "224.0.0.1",
                    "224.0.0.251",
                    "224.0.0.252",
                    "230.230.230.230",
                    "239.0.0.250",
                    "239.255.255.250",
                ],
                "ipv6_multicast": [
                    "ff01::1",
                    "ff02::1",
                    "ff02::c",
                    "ff02::fb",
                    "ff02::1:3",
                    "ff02::1:ff0f:4c48",
                    "ff02::1:ffa6:f6e6",
                ],
            }
        )
        self.mock_anycast = MagicMock(
            return_value={"ip_anycast": [], "ipv6_anycast": []}
        )
        self.mock_wins = MagicMock(return_value={"ip_wins": []})

    def test_get_interface_info_dot_net(self):
        expected = {
            "Ethernet": {
                "alias": "Ethernet",
                "description": "Dell GigabitEthernet",
                "dns_enabled": False,
                "dns_suffix": "",
                "dynamic_dns_enabled": False,
                "id": "{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
                "ip_addresses": [
                    {
                        "address": "172.18.87.49",
                        "broadcast": "172.18.87.63",
                        "loopback": "127.0.0.1",
                        "netmask": "255.255.255.240",
                        "prefix_length": 28,
                        "prefix_origin": "Manual",
                        "suffix_origin": "Manual",
                    }
                ],
                "ip_anycast": [],
                "ip_dns": ["10.4.0.1", "10.1.0.1", "8.8.8.8"],
                "ip_gateways": ["192.168.0.1"],
                "ip_multicast": [
                    "224.0.0.1",
                    "224.0.0.251",
                    "224.0.0.252",
                    "230.230.230.230",
                    "239.0.0.250",
                    "239.255.255.250",
                ],
                "ip_wins": [],
                "ipv6_addresses": [
                    {
                        "address": "fe80::e8a4:1224:5548:2b81",
                        "interface_index": 32,
                        "prefix_length": 64,
                        "prefix_origin": "WellKnown",
                        "suffix_origin": "Router",
                    }
                ],
                "ipv6_anycast": [],
                "ipv6_dns": ["2600:740a:1:304::1"],
                "ipv6_gateways": ["fe80::208:a2ff:fe0b:de70"],
                "ipv6_multicast": [
                    "ff01::1",
                    "ff02::1",
                    "ff02::c",
                    "ff02::fb",
                    "ff02::1:3",
                    "ff02::1:ff0f:4c48",
                    "ff02::1:ffa6:f6e6",
                ],
                "physical_address": "02:D5:F1:DD:31:E0",
                "receive_only": False,
                "status": "Up",
                "type": "Ethernet",
            }
        }

        mock_int = MagicMock(return_value=[Interface()])
        with patch.object(
            win_network, "_get_network_interfaces", mock_int
        ), patch.object(
            win_network, "_get_ip_base_properties", self.mock_ip_base
        ), patch.object(
            win_network, "_get_ip_unicast_info", self.mock_unicast
        ), patch.object(
            win_network, "_get_ip_gateway_info", self.mock_gateway
        ), patch.object(
            win_network, "_get_ip_dns_info", self.mock_dns
        ), patch.object(
            win_network, "_get_ip_multicast_info", self.mock_multicast
        ), patch.object(
            win_network, "_get_ip_anycast_info", self.mock_anycast
        ), patch.object(
            win_network, "_get_ip_wins_info", self.mock_wins
        ):

            # ret = win_network._get_base_properties()
            results = win_network.get_interface_info_dot_net()

        self.assertDictEqual(expected, results)

    def test_get_network_info(self):
        expected = {
            "Dell GigabitEthernet": {
                "hwaddr": "02:D5:F1:DD:31:E0",
                "inet": [
                    {
                        "address": "172.18.87.49",
                        "broadcast": "172.18.87.63",
                        "gateway": "192.168.0.1",
                        "label": "Dell GigabitEthernet",
                        "netmask": "255.255.255.240",
                    }
                ],
                "inet6": [
                    {
                        "address": "fe80::e8a4:1224:5548:2b81",
                        "gateway": "fe80::208:a2ff:fe0b:de70",
                    }
                ],
                "up": True,
            }
        }
        mock_int = MagicMock(return_value=[Interface()])
        with patch.object(
            win_network, "_get_network_interfaces", mock_int
        ), patch.object(
            win_network, "_get_ip_base_properties", self.mock_ip_base
        ), patch.object(
            win_network, "_get_ip_unicast_info", self.mock_unicast
        ), patch.object(
            win_network, "_get_ip_gateway_info", self.mock_gateway
        ), patch.object(
            win_network, "_get_ip_dns_info", self.mock_dns
        ), patch.object(
            win_network, "_get_ip_multicast_info", self.mock_multicast
        ), patch.object(
            win_network, "_get_ip_anycast_info", self.mock_anycast
        ), patch.object(
            win_network, "_get_ip_wins_info", self.mock_wins
        ):

            # ret = win_network._get_base_properties()
            results = win_network.get_interface_info()

        self.assertDictEqual(expected, results)

    def test__get_base_properties_tap_adapter(self):
        """
        Adapter Type 53 is apparently an undocumented type corresponding to
        OpenVPN TAP Adapters and possibly other TAP Adapters. This test makes
        sure the win_network util will catch that.
        https://github.com/saltstack/salt/issues/56196
        https://github.com/saltstack/salt/issues/56275
        """
        i_face = Interface(
            i_address="03DE4D0713FA",
            i_description="Windows TAP Adapter",
            i_id="{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
            i_name="Windows TAP Adapter",
            i_receive_only=False,
            i_status=1,
            i_type=53,
        )
        expected = {
            "alias": "Windows TAP Adapter",
            "description": "Windows TAP Adapter",
            "id": "{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
            "receive_only": False,
            "physical_address": "03:DE:4D:07:13:FA",
            "status": "Up",
            "type": "TAPAdapter",
        }
        results = win_network._get_base_properties(i_face=i_face)
        self.assertDictEqual(expected, results)

    def test__get_base_properties_undefined_adapter(self):
        """
        The Adapter Type 53 may be an arbitrary number assigned by OpenVPN.
        This will test the ability to avoid stack tracing on an undefined
        adapter type. If one is encountered, just use the description.
        """
        i_face = Interface(
            i_address="03DE4D0713FA",
            i_description="Undefined Adapter",
            i_id="{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
            i_name="Undefined",
            i_receive_only=False,
            i_status=1,
            i_type=50,
        )
        expected = {
            "alias": "Undefined",
            "description": "Undefined Adapter",
            "id": "{C5F468C0-DD5F-4C2B-939F-A411DCB5DE16}",
            "receive_only": False,
            "physical_address": "03:DE:4D:07:13:FA",
            "status": "Up",
            "type": "Undefined Adapter",
        }
        results = win_network._get_base_properties(i_face=i_face)
        self.assertDictEqual(expected, results)



"""
Functions to translate input for container creation
"""

import os

from salt.exceptions import SaltInvocationError

from . import helpers

ALIASES = {
    "cmd": "command",
    "cpuset": "cpuset_cpus",
    "dns_option": "dns_opt",
    "env": "environment",
    "expose": "ports",
    "interactive": "stdin_open",
    "ipc": "ipc_mode",
    "label": "labels",
    "memory": "mem_limit",
    "memory_swap": "memswap_limit",
    "publish": "port_bindings",
    "publish_all": "publish_all_ports",
    "restart": "restart_policy",
    "rm": "auto_remove",
    "sysctl": "sysctls",
    "security_opts": "security_opt",
    "ulimit": "ulimits",
    "user_ns_mode": "userns_mode",
    "volume": "volumes",
    "workdir": "working_dir",
}
ALIASES_REVMAP = {y: x for x, y in ALIASES.items()}


def _merge_keys(kwargs):
    """
    The log_config is a mixture of the CLI options --log-driver and --log-opt
    (which we support in Salt as log_driver and log_opt, respectively), but it
    must be submitted to the host config in the format {'Type': log_driver,
    'Config': log_opt}. So, we need to construct this argument to be passed to
    the API from those two arguments.
    """
    log_driver = kwargs.pop("log_driver", helpers.NOTSET)
    log_opt = kwargs.pop("log_opt", helpers.NOTSET)
    if "log_config" not in kwargs:
        if log_driver is not helpers.NOTSET or log_opt is not helpers.NOTSET:
            kwargs["log_config"] = {
                "Type": log_driver if log_driver is not helpers.NOTSET else "none",
                "Config": log_opt if log_opt is not helpers.NOTSET else {},
            }


def _post_processing(kwargs, skip_translate, invalid):
    """
    Additional container-specific post-translation processing
    """
    # Don't allow conflicting options to be set
    if kwargs.get("port_bindings") is not None and kwargs.get("publish_all_ports"):
        kwargs.pop("port_bindings")
        invalid["port_bindings"] = "Cannot be used when publish_all_ports=True"
    if kwargs.get("hostname") is not None and kwargs.get("network_mode") == "host":
        kwargs.pop("hostname")
        invalid["hostname"] = "Cannot be used when network_mode=True"

    # Make sure volumes and ports are defined to match the binds and port_bindings
    if kwargs.get("binds") is not None and (
        skip_translate is True
        or all(x not in skip_translate for x in ("binds", "volume", "volumes"))
    ):
        # Make sure that all volumes defined in "binds" are included in the
        # "volumes" param.
        auto_volumes = []
        if isinstance(kwargs["binds"], dict):
            for val in kwargs["binds"].values():
                try:
                    if "bind" in val:
                        auto_volumes.append(val["bind"])
                except TypeError:
                    continue
        else:
            if isinstance(kwargs["binds"], list):
                auto_volume_defs = kwargs["binds"]
            else:
                try:
                    auto_volume_defs = helpers.split(kwargs["binds"])
                except AttributeError:
                    auto_volume_defs = []
            for val in auto_volume_defs:
                try:
                    auto_volumes.append(helpers.split(val, ":")[1])
                except IndexError:
                    continue
        if auto_volumes:
            actual_volumes = kwargs.setdefault("volumes", [])
            actual_volumes.extend([x for x in auto_volumes if x not in actual_volumes])
            # Sort list to make unit tests more reliable
            actual_volumes.sort()

    if kwargs.get("port_bindings") is not None and all(
        x not in skip_translate for x in ("port_bindings", "expose", "ports")
    ):
        # Make sure that all ports defined in "port_bindings" are included in
        # the "ports" param.
        ports_to_bind = list(kwargs["port_bindings"])
        if ports_to_bind:
            ports_to_open = set(kwargs.get("ports", []))
            ports_to_open.update([helpers.get_port_def(x) for x in ports_to_bind])
            kwargs["ports"] = list(ports_to_open)

    if "ports" in kwargs and all(x not in skip_translate for x in ("expose", "ports")):
        # TCP ports should only be passed as the port number. Normalize the
        # input so a port definition of 80/tcp becomes just 80 instead of
        # (80, 'tcp').
        for index, _ in enumerate(kwargs["ports"]):
            try:
                if kwargs["ports"][index][1] == "tcp":
                    kwargs["ports"][index] = ports_to_open[index][0]
            except TypeError:
                continue


# Functions below must match names of docker-py arguments
def auto_remove(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def binds(val, **kwargs):  # pylint: disable=unused-argument
    """
    On the CLI, these are passed as multiple instances of a given CLI option.
    In Salt, we accept these as a comma-delimited list but the API expects a
    Python list.
    """
    if not isinstance(val, dict):
        if not isinstance(val, list):
            try:
                val = helpers.split(val)
            except AttributeError:
                raise SaltInvocationError(
                    "'{}' is not a dictionary or list of bind definitions".format(val)
                )
    return val


def blkio_weight(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_int(val)


def blkio_weight_device(val, **kwargs):  # pylint: disable=unused-argument
    """
    CLI input is a list of PATH:WEIGHT pairs, but the API expects a list of
    dictionaries in the format [{'Path': path, 'Weight': weight}]
    """
    val = helpers.map_vals(val, "Path", "Weight")
    for item in val:
        try:
            item["Weight"] = int(item["Weight"])
        except (TypeError, ValueError):
            raise SaltInvocationError(
                "Weight '{Weight}' for path '{Path}' is not an integer".format(**item)
            )
    return val


def cap_add(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_stringlist(val)


def cap_drop(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_stringlist(val)


def command(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_command(val)


def cpuset_cpus(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def cpuset_mems(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def cpu_group(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_int(val)


def cpu_period(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_int(val)


def cpu_shares(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_int(val)


def detach(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def device_read_bps(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_device_rates(val, numeric_rate=False)


def device_read_iops(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_device_rates(val, numeric_rate=True)


def device_write_bps(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_device_rates(val, numeric_rate=False)


def device_write_iops(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_device_rates(val, numeric_rate=True)


def devices(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_stringlist(val)


def dns_opt(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_stringlist(val)


def dns_search(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_stringlist(val)


def dns(val, **kwargs):
    val = helpers.translate_stringlist(val)
    if kwargs.get("validate_ip_addrs", True):
        for item in val:
            helpers.validate_ip(item)
    return val


def domainname(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def entrypoint(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_command(val)


def environment(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_key_val(val, delimiter="=")


def extra_hosts(val, **kwargs):
    val = helpers.translate_key_val(val, delimiter=":")
    if kwargs.get("validate_ip_addrs", True):
        for key in val:
            helpers.validate_ip(val[key])
    return val


def group_add(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_stringlist(val)


def host_config(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_dict(val)


def hostname(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def ipc_mode(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def isolation(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def labels(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_labels(val)


def links(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_key_val(val, delimiter=":")


def log_driver(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def log_opt(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_key_val(val, delimiter="=")


def lxc_conf(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_key_val(val, delimiter="=")


def mac_address(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def mem_limit(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bytes(val)


def mem_swappiness(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_int(val)


def memswap_limit(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bytes(val)


def name(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def network_disabled(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def network_mode(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def oom_kill_disable(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def oom_score_adj(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_int(val)


def pid_mode(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def pids_limit(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_int(val)


def port_bindings(val, **kwargs):
    """
    On the CLI, these are passed as multiple instances of a given CLI option.
    In Salt, we accept these as a comma-delimited list but the API expects a
    Python dictionary mapping ports to their bindings. The format the API
    expects is complicated depending on whether or not the external port maps
    to a different internal port, or if the port binding is for UDP instead of
    TCP (the default). For reference, see the "Port bindings" section in the
    docker-py documentation at the following URL:
    http://docker-py.readthedocs.io/en/stable/api.html
    """
    validate_ip_addrs = kwargs.get("validate_ip_addrs", True)
    if not isinstance(val, dict):
        if not isinstance(val, list):
            try:
                val = helpers.split(val)
            except AttributeError:
                val = helpers.split(str(val))

        for idx, item in enumerate(val):
            if not isinstance(item, str):
                val[idx] = str(item)

        def _format_port(port_num, proto):
            return str(port_num) + "/udp" if proto.lower() == "udp" else port_num

        bindings = {}
        for binding in val:
            bind_parts = helpers.split(binding, ":")
            num_bind_parts = len(bind_parts)
            if num_bind_parts == 1:
                # Single port or port range being passed through (no
                # special mapping)
                container_port = str(bind_parts[0])
                if container_port == "":
                    raise SaltInvocationError("Empty port binding definition found")
                container_port, _, proto = container_port.partition("/")
                try:
                    start, end = helpers.get_port_range(container_port)
                except ValueError as exc:
                    # Using __str__() to avoid deprecation warning for using
                    # the message attribute of the ValueError.
                    raise SaltInvocationError(exc.__str__())
                bind_vals = [
                    (_format_port(port_num, proto), None)
                    for port_num in range(start, end + 1)
                ]
            elif num_bind_parts == 2:
                if bind_parts[0] == "":
                    raise SaltInvocationError(
                        "Empty host port in port binding definition '{}'".format(
                            binding
                        )
                    )
                if bind_parts[1] == "":
                    raise SaltInvocationError(
                        "Empty container port in port binding definition '{}'".format(
                            binding
                        )
                    )
                container_port, _, proto = bind_parts[1].partition("/")
                try:
                    cport_start, cport_end = helpers.get_port_range(container_port)
                    hport_start, hport_end = helpers.get_port_range(bind_parts[0])
                except ValueError as exc:
                    # Using __str__() to avoid deprecation warning for
                    # using the message attribute of the ValueError.
                    raise SaltInvocationError(exc.__str__())
                if (hport_end - hport_start) != (cport_end - cport_start):
                    # Port range is mismatched
                    raise SaltInvocationError(
                        "Host port range ({}) does not have the same "
                        "number of ports as the container port range "
                        "({})".format(bind_parts[0], container_port)
                    )
                cport_list = list(range(cport_start, cport_end + 1))
                hport_list = list(range(hport_start, hport_end + 1))
                bind_vals = [
                    (_format_port(item, proto), hport_list[ind])
                    for ind, item in enumerate(cport_list)
                ]
            elif num_bind_parts == 3:
                host_ip, host_port = bind_parts[0:2]
                if validate_ip_addrs:
                    helpers.validate_ip(host_ip)
                container_port, _, proto = bind_parts[2].partition("/")
                try:
                    cport_start, cport_end = helpers.get_port_range(container_port)
                except ValueError as exc:
                    # Using __str__() to avoid deprecation warning for
                    # using the message attribute of the ValueError.
                    raise SaltInvocationError(exc.__str__())
                cport_list = list(range(cport_start, cport_end + 1))
                if host_port == "":
                    hport_list = [None] * len(cport_list)
                else:
                    try:
                        hport_start, hport_end = helpers.get_port_range(host_port)
                    except ValueError as exc:
                        # Using __str__() to avoid deprecation warning for
                        # using the message attribute of the ValueError.
                        raise SaltInvocationError(exc.__str__())
                    hport_list = list(range(hport_start, hport_end + 1))

                    if (hport_end - hport_start) != (cport_end - cport_start):
                        # Port range is mismatched
                        raise SaltInvocationError(
                            "Host port range ({}) does not have the same "
                            "number of ports as the container port range "
                            "({})".format(host_port, container_port)
                        )

                bind_vals = [
                    (
                        _format_port(val, proto),
                        (host_ip,)
                        if hport_list[idx] is None
                        else (host_ip, hport_list[idx]),
                    )
                    for idx, val in enumerate(cport_list)
                ]
            else:
                raise SaltInvocationError(
                    "'{}' is an invalid port binding definition (at most "
                    "3 components are allowed, found {})".format(
                        binding, num_bind_parts
                    )
                )

            for cport, bind_def in bind_vals:
                if cport not in bindings:
                    bindings[cport] = bind_def
                else:
                    if isinstance(bindings[cport], list):
                        # Append to existing list of bindings for this
                        # container port.
                        bindings[cport].append(bind_def)
                    else:
                        bindings[cport] = [bindings[cport], bind_def]
                    for idx, val in enumerate(bindings[cport]):
                        if val is None:
                            # Now that we are adding multiple
                            # bindings
                            try:
                                # Convert 1234/udp to 1234
                                bindings[cport][idx] = int(cport.split("/")[0])
                            except AttributeError:
                                # Port was tcp, the AttributeError
                                # signifies that the split failed
                                # because the port number was
                                # already defined as an integer.
                                # Just use the cport.
                                bindings[cport][idx] = cport
        val = bindings
    return val


def ports(val, **kwargs):  # pylint: disable=unused-argument
    """
    Like cap_add, cap_drop, etc., this option can be specified multiple times,
    and each time can be a port number or port range. Ultimately, the API
    expects a list, but elements in the list are ints when the port is TCP, and
    a tuple (port_num, 'udp') when the port is UDP.
    """
    if not isinstance(val, list):
        try:
            val = helpers.split(val)
        except AttributeError:
            if isinstance(val, int):
                val = [val]
            else:
                raise SaltInvocationError(
                    "'{}' is not a valid port definition".format(val)
                )
    new_ports = set()
    for item in val:
        if isinstance(item, int):
            new_ports.add(item)
            continue
        try:
            item, _, proto = item.partition("/")
        except AttributeError:
            raise SaltInvocationError(
                "'{}' is not a valid port definition".format(item)
            )
        try:
            range_start, range_end = helpers.get_port_range(item)
        except ValueError as exc:
            # Using __str__() to avoid deprecation warning for using
            # the "message" attribute of the ValueError.
            raise SaltInvocationError(exc.__str__())
        new_ports.update(
            [helpers.get_port_def(x, proto) for x in range(range_start, range_end + 1)]
        )
    return list(new_ports)


def privileged(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def publish_all_ports(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def read_only(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def restart_policy(val, **kwargs):  # pylint: disable=unused-argument
    """
    CLI input is in the format NAME[:RETRY_COUNT] but the API expects {'Name':
    name, 'MaximumRetryCount': retry_count}. We will use the 'fill' kwarg here
    to make sure the mapped result uses '0' for the count if this optional
    value was omitted.
    """
    val = helpers.map_vals(val, "Name", "MaximumRetryCount", fill="0")
    # map_vals() converts the input into a list of dicts, but the API
    # wants just a dict, so extract the value from the single-element
    # list. If there was more than one element in the list, then
    # invalid input was passed (i.e. a comma-separated list, when what
    # we wanted was a single value).
    if len(val) != 1:
        raise SaltInvocationError("Only one policy is permitted")
    val = val[0]
    try:
        # The count needs to be an integer
        val["MaximumRetryCount"] = int(val["MaximumRetryCount"])
    except (TypeError, ValueError):
        # Non-numeric retry count passed
        raise SaltInvocationError(
            "Retry count '{}' is non-numeric".format(val["MaximumRetryCount"])
        )
    return val


def security_opt(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_stringlist(val)


def shm_size(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bytes(val)


def stdin_open(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def stop_signal(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def stop_timeout(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_int(val)


def storage_opt(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_key_val(val, delimiter="=")


def sysctls(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_key_val(val, delimiter="=")


def tmpfs(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_dict(val)


def tty(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_bool(val)


def ulimits(val, **kwargs):  # pylint: disable=unused-argument
    val = helpers.translate_stringlist(val)
    for idx, item in enumerate(val):
        if not isinstance(item, dict):
            try:
                ulimit_name, limits = helpers.split(item, "=", 1)
                comps = helpers.split(limits, ":", 1)
            except (AttributeError, ValueError):
                raise SaltInvocationError(
                    "Ulimit definition '{}' is not in the format "
                    "type=soft_limit[:hard_limit]".format(item)
                )
            if len(comps) == 1:
                comps *= 2
            soft_limit, hard_limit = comps
            try:
                val[idx] = {
                    "Name": ulimit_name,
                    "Soft": int(soft_limit),
                    "Hard": int(hard_limit),
                }
            except (TypeError, ValueError):
                raise SaltInvocationError(
                    "Limit '{}' contains non-numeric value(s)".format(item)
                )
    return val


def user(val, **kwargs):  # pylint: disable=unused-argument
    """
    This can be either a string or a numeric uid
    """
    if not isinstance(val, int):
        # Try to convert to integer. This will fail if the value is a
        # username. This is OK, as we check below to make sure that the
        # value is either a string or integer. Trying to convert to an
        # integer first though will allow us to catch the edge case in
        # which a quoted uid is passed (e.g. '1000').
        try:
            val = int(val)
        except (TypeError, ValueError):
            pass
    if not isinstance(val, (int, str)):
        raise SaltInvocationError("Value must be a username or uid")
    elif isinstance(val, int) and val < 0:
        raise SaltInvocationError("'{}' is an invalid uid".format(val))
    return val


def userns_mode(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def volume_driver(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_str(val)


def volumes(val, **kwargs):  # pylint: disable=unused-argument
    """
    Should be a list of absolute paths
    """
    val = helpers.translate_stringlist(val)
    for item in val:
        if not os.path.isabs(item):
            raise SaltInvocationError("'{}' is not an absolute path".format(item))
    return val


def volumes_from(val, **kwargs):  # pylint: disable=unused-argument
    return helpers.translate_stringlist(val)


def working_dir(val, **kwargs):  # pylint: disable=unused-argument
    """
    Must be an absolute path
    """
    try:
        is_abs = os.path.isabs(val)
    except AttributeError:
        is_abs = False
    if not is_abs:
        raise SaltInvocationError("'{}' is not an absolute path".format(val))
    return val



"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""


import salt.modules.napalm_probes as napalm_probes
import tests.support.napalm as napalm_test_support
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class NapalmProbesModuleTestCase(TestCase, LoaderModuleMockMixin):
    @classmethod
    def setUpClass(cls):
        cls._test_probes = {
            "new_probe": {
                "new_test1": {
                    "probe_type": "icmp-ping",
                    "target": "192.168.0.1",
                    "source": "192.168.0.2",
                    "probe_count": 13,
                    "test_interval": 3,
                }
            }
        }
        cls._test_delete_probes = {
            "existing_probe": {"existing_test1": {}, "existing_test2": {}}
        }
        cls._test_schedule_probes = {
            "test_probe": {"existing_test1": {}, "existing_test2": {}}
        }

    @classmethod
    def tearDownClass(cls):
        cls._test_probes = cls._test_delete_probes = cls._test_schedule_probes = None

    def setup_loader_modules(self):
        patcher = patch(
            "salt.utils.napalm.get_device",
            MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        def mock_net_load(template, *args, **kwargs):
            if template == "set_probes":
                assert kwargs["probes"] == self._test_probes.copy()
                return napalm_test_support.TEST_TERM_CONFIG.copy()
            if template == "delete_probes":
                assert kwargs["probes"] == self._test_delete_probes.copy()
                return napalm_test_support.TEST_TERM_CONFIG.copy()
            if template == "schedule_probes":
                assert kwargs["probes"] == self._test_schedule_probes.copy()
                return napalm_test_support.TEST_TERM_CONFIG.copy()
            raise ValueError("incorrect template {}".format(template))

        module_globals = {
            "__salt__": {
                "config.get": MagicMock(
                    return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
                ),
                "file.file_exists": napalm_test_support.true,
                "file.join": napalm_test_support.join,
                "file.get_managed": napalm_test_support.get_managed_file,
                "random.hash": napalm_test_support.random_hash,
                "net.load_template": mock_net_load,
            }
        }

        return {napalm_probes: module_globals}

    def test_probes_config(self):
        ret = napalm_probes.config()
        assert ret["out"] == napalm_test_support.TEST_PROBES_CONFIG.copy()

    def test_probes_results(self):
        ret = napalm_probes.results()
        assert ret["out"] == napalm_test_support.TEST_PROBES_RESULTS.copy()

    def test_set_probes(self):
        ret = napalm_probes.set_probes(self._test_probes.copy())
        assert ret["result"] is True

    def test_delete_probes(self):
        ret = napalm_probes.delete_probes(self._test_delete_probes.copy())
        assert ret["result"] is True

    def test_schedule_probes(self):
        ret = napalm_probes.schedule_probes(self._test_schedule_probes.copy())
        assert ret["result"] is True



"""
integration tests for mac_shadow
"""

import datetime

import pytest
from tests.support.case import ModuleCase
from tests.support.helpers import random_string, runs_on

TEST_USER = random_string("RS-", lowercase=False)
NO_USER = random_string("RS-", lowercase=False)


@runs_on(kernel="Darwin")
@pytest.mark.skip_if_binaries_missing("dscl", "pwpolicy")
@pytest.mark.skip_if_not_root
class MacShadowModuleTest(ModuleCase):
    """
    Validate the mac_shadow module
    """

    def setUp(self):
        """
        Get current settings
        """
        self.run_function("user.add", [TEST_USER])

    def tearDown(self):
        """
        Reset to original settings
        """
        self.run_function("user.delete", [TEST_USER])

    @pytest.mark.slow_test
    def test_info(self):
        """
        Test shadow.info
        """
        # Correct Functionality
        ret = self.run_function("shadow.info", [TEST_USER])
        self.assertEqual(ret["name"], TEST_USER)

        # User does not exist
        ret = self.run_function("shadow.info", [NO_USER])
        self.assertEqual(ret["name"], "")

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_get_account_created(self):
        """
        Test shadow.get_account_created
        """
        # Correct Functionality
        text_date = self.run_function("shadow.get_account_created", [TEST_USER])
        self.assertNotEqual(text_date, "Invalid Timestamp")
        obj_date = datetime.datetime.strptime(text_date, "%Y-%m-%d %H:%M:%S")
        self.assertIsInstance(obj_date, datetime.date)

        # User does not exist
        self.assertEqual(
            self.run_function("shadow.get_account_created", [NO_USER]),
            "ERROR: User not found: {}".format(NO_USER),
        )

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_get_last_change(self):
        """
        Test shadow.get_last_change
        """
        # Correct Functionality
        text_date = self.run_function("shadow.get_last_change", [TEST_USER])
        self.assertNotEqual(text_date, "Invalid Timestamp")
        obj_date = datetime.datetime.strptime(text_date, "%Y-%m-%d %H:%M:%S")
        self.assertIsInstance(obj_date, datetime.date)

        # User does not exist
        self.assertEqual(
            self.run_function("shadow.get_last_change", [NO_USER]),
            "ERROR: User not found: {}".format(NO_USER),
        )

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_get_login_failed_last(self):
        """
        Test shadow.get_login_failed_last
        """
        # Correct Functionality
        text_date = self.run_function("shadow.get_login_failed_last", [TEST_USER])
        self.assertNotEqual(text_date, "Invalid Timestamp")
        obj_date = datetime.datetime.strptime(text_date, "%Y-%m-%d %H:%M:%S")
        self.assertIsInstance(obj_date, datetime.date)

        # User does not exist
        self.assertEqual(
            self.run_function("shadow.get_login_failed_last", [NO_USER]),
            "ERROR: User not found: {}".format(NO_USER),
        )

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_get_login_failed_count(self):
        """
        Test shadow.get_login_failed_count
        """
        # Correct Functionality
        self.assertEqual(
            self.run_function("shadow.get_login_failed_count", [TEST_USER]), "0"
        )

        # User does not exist
        self.assertEqual(
            self.run_function("shadow.get_login_failed_count", [NO_USER]),
            "ERROR: User not found: {}".format(NO_USER),
        )

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_get_set_maxdays(self):
        """
        Test shadow.get_maxdays
        Test shadow.set_maxdays
        """
        # Correct Functionality
        self.assertTrue(self.run_function("shadow.set_maxdays", [TEST_USER, 20]))
        self.assertEqual(self.run_function("shadow.get_maxdays", [TEST_USER]), 20)

        # User does not exist
        self.assertEqual(
            self.run_function("shadow.set_maxdays", [NO_USER, 7]),
            "ERROR: User not found: {}".format(NO_USER),
        )
        self.assertEqual(
            self.run_function("shadow.get_maxdays", [NO_USER]),
            "ERROR: User not found: {}".format(NO_USER),
        )

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_get_set_change(self):
        """
        Test shadow.get_change
        Test shadow.set_change
        """
        # Correct Functionality
        self.assertTrue(
            self.run_function("shadow.set_change", [TEST_USER, "02/11/2011"])
        )
        self.assertEqual(
            self.run_function("shadow.get_change", [TEST_USER]), "02/11/2011"
        )

        # User does not exist
        self.assertEqual(
            self.run_function("shadow.set_change", [NO_USER, "02/11/2012"]),
            "ERROR: User not found: {}".format(NO_USER),
        )
        self.assertEqual(
            self.run_function("shadow.get_change", [NO_USER]),
            "ERROR: User not found: {}".format(NO_USER),
        )

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_get_set_expire(self):
        """
        Test shadow.get_expire
        Test shadow.set_expire
        """
        # Correct Functionality
        self.assertTrue(
            self.run_function("shadow.set_expire", [TEST_USER, "02/11/2011"])
        )
        self.assertEqual(
            self.run_function("shadow.get_expire", [TEST_USER]), "02/11/2011"
        )

        # User does not exist
        self.assertEqual(
            self.run_function("shadow.set_expire", [NO_USER, "02/11/2012"]),
            "ERROR: User not found: {}".format(NO_USER),
        )
        self.assertEqual(
            self.run_function("shadow.get_expire", [NO_USER]),
            "ERROR: User not found: {}".format(NO_USER),
        )

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_del_password(self):
        """
        Test shadow.del_password
        """
        # Correct Functionality
        self.assertTrue(self.run_function("shadow.del_password", [TEST_USER]))
        self.assertEqual(self.run_function("shadow.info", [TEST_USER])["passwd"], "*")

        # User does not exist
        self.assertEqual(
            self.run_function("shadow.del_password", [NO_USER]),
            "ERROR: User not found: {}".format(NO_USER),
        )

    @pytest.mark.destructive_test
    @pytest.mark.slow_test
    def test_set_password(self):
        """
        Test shadow.set_password
        """
        # Correct Functionality
        self.assertTrue(
            self.run_function("shadow.set_password", [TEST_USER, "Pa$$W0rd"])
        )

        # User does not exist
        self.assertEqual(
            self.run_function("shadow.set_password", [NO_USER, "P@SSw0rd"]),
            "ERROR: User not found: {}".format(NO_USER),
        )
